#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Common Utility functions for convinience - no dependencies from other moduls  
"""

import os
import json
from pathlib                import Path
from termcolor              import colored
from platformdirs           import user_config_dir, user_data_dir

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)



#------------------------------------------------------------------------------
# base type utils  
#------------------------------------------------------------------------------


def clip(val, min_, max_):
    """ clip aVal to be between min and max"""
    return min_ if val < min_ else max_ if val > max_ else val



#------------------------------------------------------------------------------
# logging 
#------------------------------------------------------------------------------

def init_logging (level= logging.WARNING):
    """ initialize logging with level"""

    ch = logging.StreamHandler()

    ch.setFormatter(CustomFormatter())

    logging.basicConfig(format='%(levelname)-8s- %(message)s', 
                        handlers=[ch], 
                        level=level)  # DEBUG or WARNING
    # suppress debug messages from these modules 
    # logging.getLogger('PIL.PngImagePlugin').disabled = True
    # logging.getLogger('dxf_utils').disabled = True


class CustomFormatter(logging.Formatter):
    """ colored formatting of logging stream"""

    # format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG:    colored(" - %(levelname)s - %(message)s", 'yellow', attrs=["dark"]) + colored(" (%(filename)s:%(lineno)d)", 'white', attrs=["dark"]),                          # grey + format + reset,
        logging.INFO:     colored(" - %(message)s", 'white', attrs=["dark"]),
        logging.WARNING:  colored("WARNING - ", 'yellow') + "%(message)s",
        logging.ERROR:    colored("ERROR - ", 'red') + "%(message)s",
        logging.CRITICAL: colored("ERROR - ", 'red', attrs=["bold"]) + "%(message)s"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)
    

#------------------------------------------------------------------------------
# Dictonary handling
#------------------------------------------------------------------------------

def fromDict(aDict : dict, key, default='no default'):
    """
    returns a value from dict. If ky is not in the dict and there is no default value an error
    will be raised 

    Args:
        dict: the dictonary to look in 
        key: the key to look for       
        default: the value if key is missing
    """
    preferedType = None

    if default != 'no default':
        if isinstance (default, float):
           preferedType = float
        elif isinstance (default, bool):
            preferedType = bool
        elif isinstance (default, int):
            preferedType = int

    if aDict and key in aDict:
        value = aDict[key]
        if preferedType == float:
            value = float(value)
        elif preferedType == int:
            value = int(value)
        elif preferedType == bool:
            value = bool(value)
    else:
        if default == 'no default':
            value = None
            logger.error ('Mandatory parameter \'%s\' not specified'  % key)
        else:
            value = default 
            if value:
                logger.debug ('Parameter \'%s\' not specified, using default-value \'%s\'' % (key, str(value)))
    return value


def toDict(aDict : dict, key, value):
    """
    writes t0 the parameter dictionary. If 'value' is None the key is not written 
    """
    if isinstance (aDict,dict):
        if not value is None: 
            # limit decimals in file 
            if isinstance  (value, float):
                value = round (value,6)
            aDict [key] = value
        else: 
            # remove key from dictionary  - so default values will be used 
            aDict.pop(key, None)

        
#------------------------------------------------------------------------------
# Settings and Paramter file 
#------------------------------------------------------------------------------


class Parameters ():
    """ Handles a parameter file with a json structure representing a dictionary of paramteres""" 

    def __init__ (self, paramFilePath):

        self._paramFilePath = paramFilePath

    def get_dataDict (self):
        """
        returns the complete dataDict of self
        """
        dataDict = {}
        if self._paramFilePath:
            try:
                paramFile = open(self._paramFilePath)
                try:
                    dataDict = json.load(paramFile)
                    paramFile.close()
                except ValueError as e:
                    logger.error ("Invalid json expression '%s' in parameter file '%s'" % (e, self._paramFilePath))
                    paramFile.close()
                    dataDict = {}
            except:
                logger.debug (f"Paramter file {self._paramFilePath} not found")

        return dataDict


    def write_dataDict (self, aDict, dataName='Parameters'):
        """ writes data dict to file - returns True if succeded"""

        ok = False

        with open(self._paramFilePath, 'w+') as file:

            # save parameter dictionary to .json-file
            try:
                json.dump(aDict, file, indent=2, separators=(',', ':'))
                logger.info (f"{dataName} saved to {self._paramFilePath}" )
                ok =  True

            except ValueError as e:
                logger.error (f"Invalid json expression '{e}'. Failed to save data to '{self._paramFilePath}'")

            except TypeError as e:
                logger.error (f"{e}. Failed to save data to '{self._paramFilePath}'")

        return ok 


class Settings (Parameters):
    """ Handles a named setting file with a json structure representing a dictionary of paramteres""" 

    settingsFilePath = None                     # the filePath of the settings

    def __init__ (self):
        """ 
        object to handle i/o of settings parameters 
        
        Settings file is defined with class method 'belongsTo'
        """

        super().__init__(self.settingsFilePath)


    @staticmethod
    def user_data_dir (app_name, app_author = 'jxjo'):
        """ returns directory for app data 
        
        Args:
            app_name: name of app self will belong to 
            app_author : typical for Windows - something like 'Microsoft' of 'jxjo'  
        """

        return user_data_dir (app_name, app_author, ensure_exists=True) 

    @classmethod
    def set_path (cls, app_name, app_author = 'jxjo', name_suffix=None, file_extension= '.json'):
        """ static set of the file the settings will belong to 
        
        Args:
            app_name: name of app self will belong to 
            app_author : typical for Windows - something like 'Microsoft' of 'jxjo'  
            name_suffix: ... will be appended to appName - default '_settings'      
            file_extension: ... of the settings file - default 'json'       
        """

        fileName = app_name + name_suffix + file_extension if name_suffix else app_name + file_extension

        # get directory where self is located
        settings_dir  = user_config_dir (app_name, app_author, ensure_exists=True)

        cls.settingsFilePath = os.path.join(settings_dir, fileName)

        logger.info (f"Reading settings from {cls.settingsFilePath}")


    @classmethod
    def belongTo (cls, belongsToPath, nameExtension='_settings', fileExtension= '.json'):
        """ static set of the file the settings will belong to 
        
        Args:
            :belongsToPath: file path of the python module self will belong to 
            :nameExtension: ... will be appended to appName - default '_settings'       \n
            :fileExtension: ... of the settings file - default 'json'       \n
        """

        appName = os.path.splitext(os.path.basename(belongsToPath))[0]

        if nameExtension:
            paramFile = appName + nameExtension + fileExtension
        else:
            paramFile = appName  + fileExtension

        # get directory where 'belongTo' is located
        script_dir  = os.path.dirname(os.path.realpath(belongsToPath))

        cls.settingsFilePath = os.path.join(script_dir, paramFile)

        logger.info (f"Reading settings from {cls.settingsFilePath}")


    @property
    def filePath (self): return self._paramFilePath

    def get(self, key, default='no default'):
        """
        returns the value of 'key' from settings

        Args:
            :key: the key to look for       \n
            :default: the value if key is missing
        """
        dataDict = self.get_dataDict ()
        return fromDict(dataDict, key, default=default)

    def set(self, key, value):
        """
        sets 'key' with 'value' into settings

        Args:
            :key: the key to look for       
            :value: the value to set 
        """
        dataDict = self.get_dataDict ()

        toDict(dataDict, key, value)
        self.write_dataDict (dataDict)


    def write_dataDict (self, aDict, dataName='Settings'):
        """ writes data dict to file """
 
        super().write_dataDict (aDict, dataName=dataName)


#------------------------------------------------------------------------------
# File, Path handling 
#------------------------------------------------------------------------------

class PathHandler(): 
    """ handles relative Path of actual files to a workingDir """

    def __init__ (self, workingDir=None, onFile=None): 
        """  Pathhandler for working directory either from 'workinfDir' directly or based 'onFile' """
        self._workingDir = None

        if workingDir is not None: 
           self.workingDir = workingDir 
        elif onFile is not None:
           self.set_workingDirFromFile (onFile)

    @classmethod
    def relPath(cls, pathFilename, start = None):
        """Return a relative version of a path - like os.path.relpath - but checks for same drive

        Args:
            :start: start dir - default: cwd (current working Dir)
        """
        if start is None: start = os.getcwd()

        if Path(pathFilename).anchor == Path(start).anchor: 
            relPath = os.path.relpath(pathFilename, start = start)
        else: 
            relPath = pathFilename
        return relPath  


    @property 
    def workingDir (self):
        return self._workingDir if (not self._workingDir is None) else ''
    
    @property 
    def workingDir_name (self): 
        """ the directory name of workingDir """
        return os.path.basename(os.path.normpath(self.workingDir))
    
    @workingDir.setter
    def workingDir (self, newDir):

        if newDir is None: 
            self._workingDir = os.getcwd ()      # if no directory set, take current working Dir 
        elif not newDir: 
            self._workingDir = os.getcwd ()      # if no directory set, take current working Dir 
        elif not os.path.isdir(newDir): 
            os.makedirs(newDir)  
            self._workingDir = os.path.normpath(newDir)    
        else: 
            self._workingDir = os.path.normpath(newDir) 

    def set_workingDirFromFile (self, aFilePath):

        if aFilePath is None: 
            self.workingDir = None
        elif not aFilePath: 
            self.workingDir = os.getcwd ()      # if no directory set, take current working Dir 
        else: 
            self.workingDir = os.path.dirname(aFilePath) 
            if not os.path.isdir(self.workingDir):
                os.makedirs(self.workingDir)

    def relFilePath (self, aFilePath):
        """ returns the relative path of aFilePath to the workingDir of self"""
        if aFilePath is None: 
            return None
        else: 
            try: 
                relPath =  os.path.normpath(os.path.relpath(aFilePath, start = self.workingDir))
                if len(relPath) > len(aFilePath): 
                    return aFilePath                # relPath would be more complicated
                else: 
                    return relPath 
            except:                                 # aFilePath is on different drive 
                return aFilePath 
    
    def fullFilePath (self, aRelPath):
        """ returns the full path of relative aRelPath and the workingDir of self"""

        if aRelPath is None: 
            return self.workingDir
        else: 
            if os.path.isabs (aRelPath): 
                # maybe we can make a rel path out of it? 
                newPath = self.relFilePath (aRelPath)
                if os.path.isabs (newPath):
                    return aRelPath                 # we surrender - it's absolute
                else: 
                    aRelPath = newPath              # now we have a real rel path 
            return os.path.normpath(os.path.join (self.workingDir, aRelPath))





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""  

UI panels 

"""

import fnmatch   
from shutil                 import copyfile

from PyQt6.QtWidgets        import QDialog, QFileDialog

from base.widgets           import * 
from base.panels            import Edit_Panel, Toaster, MessageBox

from airfoil_dialogs        import Airfoil_Info_Dialog, Polar_Definition_Dialog
from airfoil_widgets        import Airfoil_Select_Open_Widget, mode_color

from xo2_dialogs            import *

from model.airfoil          import Airfoil
from model.polar_set        import Polar_Definition
from model.case             import Case_Optimize
from model.xo2_input        import *


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)



class Panel_Xo2_Abstract (Edit_Panel):
    """ 
    Abstract superclass for Edit/View-Panels of AirfoilEditor Optimize mode
        - has semantics of App
        - connect / handle signals 
    """
    from app import Main

    @property
    def app (self) -> Main:
        return self._app 
    
    @property
    def case (self) -> Case_Optimize:
        return self.dataObject

    @property
    def input_file (self) -> Input_File:
        return self.case.input_file

    @property
    def mode_optimize (self) -> bool:
        """ panel in optimize_mode or disabled ? - from App """
        return isinstance (self.case, Case_Optimize) 


    @override
    @property
    def _isDisabled (self) -> bool:
        """ overloaded: only enabled when not running """
        return self.case.isRunning if self.mode_optimize else False


    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if mode optimize """
        return self.mode_optimize
    

    @override
    def _set_panel_layout (self ):
        """ Set layout of self._panel """
        # overridden to connect to widgets changed signal

        super()._set_panel_layout ()
        for w in self.widgets:
            w.sig_changed.connect (self._on_widget_changed)   


    def _on_widget_changed (self, widget):
        """ user changed data in widget"""
        logger.debug (f"{self} {widget} widget changed slot")
        self.app._on_xo2_input_changed ()




class Panel_File_Optimize (Panel_Xo2_Abstract):
    """ File panel with open / save / ... """

    name = 'Optimize Mode'

    @override
    @property
    def _isDisabled (self) -> bool:
        """ overloaded: disable when optimize run dialog is open"""
        return self.app._xo2_run_dialog is not None

    @override
    def _on_widget_changed (self, *_):
        """ user changed data in widget"""
        # no automatic change handling 
        pass


    @property
    def workingDir (self) -> str:
        #todo 
        return self.app.workingDir


    def _init_layout (self): 

        self.set_background_color (**mode_color.OPTIMIZE)

        l = QGridLayout()
        r,c = 0, 0 
        ComboBox    (l,r,c, colSpan=2,  
                        get=lambda:self._input_fileName, set=self._set_input_fileName,
                        options= lambda: Input_File.files_in_dir (self.workingDir),
                        toolTip="The Xoptfoil2 input file")
        ToolButton  (l,r,c+3, icon=Icon.OPEN, set=self._open_input_file, toolTip="Select a Xoptfoil2 input file")
        r += 1
        Button      (l,r,c, width=90, text="New Version", set=self.app.case_optimize_new_version,
                        toolTip="Create a new version of the existing input file")
        Button      (l,r,c+1, width=90, text="New ...", set=self.app.optimize_new,
                        toolTip="Create a new input file")
        r += 1
        SpaceR      (l,r,height=10, stretch=0)
        r += 1
        Button      (l,r,c, text="Run Xoptfoil2", colSpan=2, button_style = button_style.PRIMARY,
                        set=self.app.optimize_open_run, toolTip="Run Optimizer Xoptfoil2")        
        r += 1
        SpaceR      (l,r)
        r += 1
        Button      (l,r,c,  text="&Close",  width=90, 
                        set=lambda : self.app.mode_optimize_finished(),
                        toolTip="Leave optimization mode")
        r += 1
        SpaceR      (l,r, height=5, stretch=0)

        l.setColumnStretch (1,2)
        l.setContentsMargins (QMargins(0, 0, 0, 0)) 

        return l


    @property
    def _input_fileName (self) -> str:
        return self.input_file.fileName

    def _set_input_fileName (self, aFileName : str, workingDir=None):
        """ set new case by name of input file"""
        if workingDir is None: 
            workingDir = self.workingDir

        self.app.optimize_change_case (aFileName, workingDir)


    def _open_input_file (self):
        """ open a new airfoil and load it"""

        # build somethinglike "*.inp *.xo2" as filter for the dialog
        filter_string = ""
        for extension in Input_File.INPUT_FILE_EXT:
            filter_string += f" *{extension}" if filter_string else f"*{extension}"

        filters  = f"Xoptfoil2 Input files ({filter_string})"

        newPathFileName, *_ = QFileDialog.getOpenFileName(self, filter=filters, directory=self.workingDir)

        if newPathFileName:                         # user pressed open

            self._set_input_fileName (os.path.basename (newPathFileName), workingDir=os.path.dirname (newPathFileName))

            self.refresh()



class Panel_Xo2_Case (Panel_Xo2_Abstract):
    """ Main panel of optimization"""

    name = 'Case'
    _width  = (320, 400)

    # @override
    # def _add_to_header_layout(self, l_head: QHBoxLayout):
    #     """ add Widgets to header layout"""

    #     Button (l_head, text="Run Xoptfoil2", width=120, button_style = button_style.PRIMARY,
    #             set=self.app.optimize_open_run, toolTip="Run Optimizer Xoptfoil2")        

    @property
    def optimization_options (self) -> Nml_optimization_options:
        return self.case.input_file.nml_optimization_options

    @property
    def info (self) -> Nml_info:
        return self.case.input_file.nml_info


    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        Label  (l,r,c,  colSpan=2, height=(None,None), style=style.COMMENT,
                        get=lambda: self.info.descriptions_string if self.info.descriptions_string else 'Enter a description ...')
        ToolButton (l,r,c+2, icon=Icon.EDIT, align=ALIGN_TOP,
                        set=self._edit_description)
        r += 1
        SpaceR (l,r, stretch=0)
        # r += 1
        # Label  (l,r,c,  style=style.COMMENT, get='Author',
        #                 hide=lambda: not self.input_file.nml_info.author)
        # Label  (l,r,c+1,style=style.COMMENT,
        #                 get=lambda: self.input_file.nml_info.author,
        #                 hide=lambda: not self.input_file.nml_info.author)
        r += 1
        Field       (l,r,c, lab="Final Airfoil", 
                     get=lambda: self.input_file.airfoil_final_fileName,
                     style=lambda: style.GOOD if self.case.airfoil_final else style.NORMAL,
                     toolTip=lambda: (self.case.airfoil_final.info_as_html if self.case.airfoil_final else "it does not yet exist"))
        # ToolButton  (l,r,c+2, icon=Icon.SHOW_INFO, set=self._show_info_airfoil_final,
        #              hide=lambda: self.case.airfoil_final is None,
        #              toolTip="Edit options of shape functions")
        r += 1
        Label       (l,r,c, get="Seed Airfoil", lab_disable=True)
        w = Airfoil_Select_Open_Widget (l,r,c+1, colSpan=2, textOpen="&Open", widthOpen=90, 
                                    obj=lambda: self.input_file, prop=Input_File.airfoil_seed)
        w.sig_changed.connect (self._airfoil_seed_changed)                  # check working dir 

        r += 1
        ComboBox    (l,r,c, lab="Shape functions", lab_disable=True,
                     obj=self.optimization_options, prop=Nml_optimization_options.shape_functions_label_long,
                     options=lambda: self.optimization_options.shape_functions_list,
                     disable=lambda: self.input_file.airfoil_seed.isBezierBased,
                     toolTip=lambda: "Bezier based seed airfoil is master of shape functions" if self.input_file.airfoil_seed.isBezierBased \
                                     else "Select shape functions for optimization")     
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self._edit_shape_functions,
                     toolTip="Edit options of shape functions")


        r += 1
        l.setRowStretch (r,1)
        l.setColumnMinimumWidth (0,90)
        l.setColumnStretch (1,3)
        return l 


    def _edit_description (self):
        """ open text editor to edit Xoptfoil2 input description"""

        dialog = Xo2_Description_Dialog (self,self.input_file.nml_info.descriptions_string, 
                                         parentPos=(0.8,0,3), dialogPos=(0,1))
        dialog.exec () 

        if dialog.result() == QDialog.DialogCode.Accepted:
            descriptions = dialog.new_text.split ("\n")
            self.input_file.nml_info.set_descriptions (descriptions)
            
            self.app._on_xo2_input_changed ()


    def _show_info_airfoil_final (self):
        """ show little info dialog about final airfoil"""

        dialog = Airfoil_Info_Dialog (self,self.case.airfoil_final, parentPos=(0.8,0,3), dialogPos=(0,1))
        dialog.exec () 


    def _edit_shape_functions (self):
        """ open editor to change the paramters of the shape functions"""
        diag = None 
        parentPos=(0.8, 0.0)
        dialogPos=(0.2,1.1)
        if self.optimization_options.shape_functions == Nml_optimization_options.BEZIER:
            diag = Xo2_Bezier_Dialog (self, getter=self.input_file.nml_bezier_options, 
                                         parentPos=parentPos, dialogPos=dialogPos)
        if self.optimization_options.shape_functions == Nml_optimization_options.HICKS_HENNE:
            diag = Xo2_Hicks_Henne_Dialog (self, getter=self.input_file.nml_hicks_henne_options, 
                                         parentPos=parentPos, dialogPos=dialogPos)
        if self.optimization_options.shape_functions == Nml_optimization_options.CAMB_THICK:
            diag = Xo2_Camb_Thick_Dialog (self, getter=self.input_file.nml_camb_thick_options, 
                                         parentPos=parentPos, dialogPos=dialogPos)

        if diag:
            diag.exec()
            self.refresh ()
            self.app._on_xo2_input_changed()


    def _airfoil_seed_changed (self, *_):
        """ slot seed airfoil changed - check if still in working dir"""

        airfoil_seed = self.input_file.airfoil_seed
        # the airfoil may not have a directory as it should be relative to its working dir 
        if airfoil_seed.pathName:

            # copy seed airfoil to working dir 
            airfoil_seed.saveAs (dir=self.input_file.workingDir, isWorkingDir=True)

            text = f"Seed airfoil <b>{airfoil_seed.fileName}</b> copied to working directory."
            MessageBox.info(self, "Copy seed airfoil", text)

            # QTimer.singleShot (10, self.refresh)




class Panel_Xo2_Operating_Conditions (Panel_Xo2_Abstract):
    """ Define operating conditions"""

    name = 'Operating Conditions'
    _width  = (230,350)


    @property
    def operating_conditions (self) -> Nml_operating_conditions:
        return self.input_file.nml_operating_conditions
    
    
    @property
    def opPoint_defs (self) -> OpPoint_Definitions:
        return self.input_file.opPoint_defs


    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        Label       (l,r,c, get="Default Polar", lab_disable=True)
        r += 1
        Field       (l,r,c, width=None, 
                    get=lambda: self.opPoint_defs.polar_def_default.name)
        ToolButton  (l,r,c+1, icon=Icon.EDIT,   set=self._edit_polar_def)
        r += 1
        CheckBox    (l,r,c, text="Dynamic Weighting", 
                    obj=lambda:  self.operating_conditions, prop=Nml_operating_conditions.dynamic_weighting)
        r += 1
        CheckBox    (l,r,c, text="Use Flaps", 
                    obj=lambda:  self.operating_conditions, prop=Nml_operating_conditions.use_flap)
        
        r += 1
        l.setRowStretch (r,3)
        l.setColumnMinimumWidth (0,80)
        l.setColumnStretch (0,1)

        return l


    def set_polar_def (self, polar_def_str: str):
        """ set defualt polar definition by string"""
        polar_defs : list [Polar_Definition] = self.app.polar_definitions
        for polar_def in polar_defs:
            if polar_def.name == polar_def_str:
                self.opPoint_defs.set_polar_def_default (polar_def)
                break


    def _edit_polar_def (self):
        """ edit default polar definition"""

        polar_def = self.opPoint_defs.polar_def_default
        diag = Polar_Definition_Dialog (self, polar_def, small_mode=True, parentPos=(0.9,0.1), dialogPos=(0,1))

        diag.setWindowTitle ("Default Polar of Op Points")
        diag.exec()

        self.opPoint_defs.set_polar_def_default (polar_def)
        self.app._on_xo2_input_changed()



class Panel_Xo2_Operating_Points (Panel_Xo2_Abstract):
    """ Define op Points of namelist operating_conditions"""

    name = 'Operating Points'
    _width  = 270

    def __init__ (self, *args, **kwargs):

        self._cur_opPoint_def = None 

        super().__init__ (*args, **kwargs)

    
    @property
    def opPoint_defs (self) -> OpPoint_Definitions:
        return self.input_file.opPoint_defs

    @property
    def cur_opPoint_def (self) -> OpPoint_Definition:
        """ current, selected opPoint def"""
        return self.opPoint_defs.current_opPoint_def


    def _init_layout (self): 

        l = QGridLayout()
        r,c = 0, 0 
        ListBox    (l,r,c, width=None, rowSpan=4, height=140,
                    get=lambda: self.cur_opPoint_def.labelLong if self.cur_opPoint_def else None ,
                    set=self.set_cur_opPoint_def_from_label,
                    signal=False,                                               # do not signal xo2_input changed
                    doubleClick=self._edit_opPoint_def,
                    options=lambda:  [opPoint_def.labelLong for opPoint_def in self.opPoint_defs])
        c += 1
        ToolButton  (l,r,c, icon=Icon.EDIT, align=ALIGN_RIGHT, 
                     set=self._edit_opPoint_def,
                     disable=lambda: not self.cur_opPoint_def)
        r += 1
        ToolButton  (l,r,c, icon=Icon.DELETE, align=ALIGN_RIGHT,
                     set=self._delete_opPoint_def,
                     disable=lambda: not self.opPoint_defs)
        r += 1
        ToolButton  (l,r,c, icon=Icon.ADD, align=ALIGN_RIGHT, 
                     set=self._add_opPoint_def)
        r += 1
        l.setRowStretch (r,3)
        l.setColumnStretch (0,2)
        l.setColumnMinimumWidth (c,30)

        return l


 
    def set_cur_opPoint_def_from_label (self, aStr: str):
        """ set from labelLong - for ComboBox"""
        for opPoint_def in self.opPoint_defs:
            if opPoint_def.labelLong == aStr:
                self.opPoint_defs.set_current_opPoint_def (opPoint_def) 
                self.app._on_xo2_opPoint_def_selected()
                break


    def _delete_opPoint_def (self):
        """ delete me opPoint_dev"""

        self.opPoint_defs.delete (self.cur_opPoint_def)

        self.refresh ()
        self.app._on_xo2_input_changed()


    def _add_opPoint_def (self):
        """ add a new opPoint_dev after current """

        self.opPoint_defs.create_after (self.cur_opPoint_def)

        self.refresh ()
        self.app._on_xo2_input_changed()


    def _edit_opPoint_def (self):
        """ open dialog to edit current opPoint def"""

        parentPos=(0.9, 0.2) 
        dialogPos=(0,1)

        self.app.edit_opPoint_def (self, parentPos, dialogPos)



class Panel_Xo2_Geometry_Targets (Panel_Xo2_Abstract):
    """ Edit geometry target """

    name = 'Geometry Targets'
    _width  = 250

    @property
    def geometry_targets (self) -> Nml_geometry_targets:
        return self.case.input_file.nml_geometry_targets

    @property
    def thickness (self) -> GeoTarget_Definition | None: 
        return self.geometry_targets.thickness

    @property
    def camber (self) -> GeoTarget_Definition | None: 
        return self.geometry_targets.camber

    def _init_layout (self) -> QGridLayout:

        l = QGridLayout()
        r,c = 0, 0
        CheckBox    (l,r,c, text="Thickness", colSpan=2, 
                     get=lambda: self.thickness is not None, 
                     set=lambda x: self.geometry_targets.activate_thickness(x))
        FieldF      (l,r,c+2, width=70, unit="%", step=0.01,
                     obj=lambda: self.thickness, prop=GeoTarget_Definition.optValue,
                     hide=lambda: not self.thickness)
        r += 1
        FieldF      (l,r,c+1, lab="Weighting", lab_disable=True, width=70, step=0.1, lim=(-10,10), dec=2,
                     obj=lambda: self.thickness, prop=GeoTarget_Definition.weighting_abs,
                     hide=lambda: self.thickness is None)
        CheckBox    (l,r,c+3, text="Fix", align=ALIGN_RIGHT,
                     obj=lambda: self.thickness, prop=OpPoint_Definition.weighting_fixed,
                     hide=lambda: self.thickness is None,
                     toolTip="Fix this weighting during Dynamic Weighting")

        r += 1
        CheckBox    (l,r,c, text="Camber", colSpan=2, 
                     get=lambda: self.camber is not None, 
                     set=lambda x: self.geometry_targets.activate_camber(x))
        FieldF      (l,r,c+2, width=70, unit="%", step=0.01,
                     obj=lambda: self.camber, prop=GeoTarget_Definition.optValue,
                     hide=lambda: not self.camber)
        r += 1
        FieldF      (l,r,c+1, lab="Weighting", lab_disable=True, width=70, step=0.1, lim=(-10,10), dec=2,
                     obj=lambda: self.camber, prop=GeoTarget_Definition.weighting_abs,
                     hide=lambda: self.camber is None)
        CheckBox    (l,r,c+3, text="Fix", align=ALIGN_RIGHT,
                     obj=lambda: self.camber, prop=OpPoint_Definition.weighting_fixed,
                     hide=lambda: self.camber is None,
                     toolTip="Fix this weighting during Dynamic Weighting")
        r += 1
        l.setRowStretch (r,2)
        r += 1
        Label (l,r,c, colSpan=4, style=style.COMMENT,
               get=lambda: f"Will be {len (self.geometry_targets.geoTarget_defs)} design variables",
               hide=lambda: len (self.geometry_targets.geoTarget_defs )== 0)        
        l.setColumnMinimumWidth (0,20)
        l.setColumnMinimumWidth (1,70)
        l.setColumnMinimumWidth (3,50)
        l.setColumnStretch (4,2)

        return l



class Panel_Xo2_Curvature (Panel_Xo2_Abstract):
    """ Edit curvature paramters """

    name = 'Curvature'
    _width  = 200


    @override
    @property
    def shouldBe_visible (self) -> bool:
        """ overloaded: only visible if mode optimize and not camb_thick """
        return super().shouldBe_visible and self.shape_functions != Nml_optimization_options.CAMB_THICK


    @property
    def shape_functions (self) -> str:
        return self.input_file.nml_optimization_options.shape_functions

    @property
    def curvature (self) -> Nml_curvature:
        return self.input_file.nml_curvature


    def _init_layout (self) -> QGridLayout:

        l = QGridLayout()
        r,c = 0, 0
        CheckBox    (l,r,c, text="Check curvature ", 
                     obj=lambda: self.curvature, prop=Nml_curvature.check_curvature)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, 
                     toolTip="Edit options",
                     set=self._edit_curvature_options,
                     hide=lambda: not self.curvature.check_curvature)
        r += 1
        Label       (l,r,c, style=style.COMMENT, colSpan=2, lab_disable=True,  
                     get="Allow reversal on ...",
                     hide=lambda: not self.curvature.check_curvature)
        r += 1
        CheckBox    (l,r,c, text="Top side (reflexed)", colSpan=2, 
                     get=lambda: self.curvature.max_curv_reverse_top == 1,
                     set=lambda x: self.curvature.set_max_curv_reverse_top(x),
                     hide=lambda: not self.curvature.check_curvature)
        r += 1
        CheckBox    (l,r,c, text="Bot side (rearloaded)", colSpan=2,   
                     get=lambda: self.curvature.max_curv_reverse_bot == 1,
                     set=lambda x: self.curvature.set_max_curv_reverse_bot(x),
                     hide=lambda: not self.curvature.check_curvature)
        r += 1
        l.setRowStretch (r,2)
        l.setColumnStretch (0,2)

        return l


    def _edit_curvature_options (self): 

        diag = Xo2_Curvature_Dialog (self, getter=self.input_file.nml_curvature,
                                     shape_functions = self.input_file.nml_optimization_options.shape_functions, 
                                     parentPos=(0.5, 0.0), dialogPos=(0.8,1.1))
        diag.exec()
        self.refresh ()
        self.app._on_xo2_input_changed()



class Panel_Xo2_Advanced (Panel_Xo2_Abstract):
    """ Panel with the advanced input options """

    name = 'Advanced'
    _width  = 210


    @override
    def _add_to_header_layout(self, l_head: QHBoxLayout):
        """ add Widgets to header layout"""

        Button (l_head, text="Input File", width=70, button_style = button_style.SUPTLE,
                set=self._edit_input_file, toolTip="Direct editing of the Xoptfoil2 input file")        


    def _init_layout (self) -> QGridLayout:

        l = QGridLayout()
        r,c = 0, 0

        CheckBox    (l,r,c,   get=lambda: not self.input_file.nml_particle_swarm_options.isDefault)
        Label       (l,r,c+1, get="Particle Swarm Options", lab_disable=True)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self._edit_particle_swarm_options,
                     toolTip="Edit options")
        r += 1
        CheckBox    (l,r,c,   get=lambda: not self.input_file.nml_xfoil_run_options.isDefault)
        Label       (l,r,c+1, get="Xfoil Options",lab_disable=True)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self._edit_xfoil_run_options,
                     toolTip="Edit options")
        r += 1
        CheckBox    (l,r,c,   get=lambda: not self.input_file.nml_paneling_options.isDefault)
        Label       (l,r,c+1, get="Paneling Options", lab_disable=True)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self._edit_paneling_options,
                     toolTip="Edit options")

        r += 1
        CheckBox    (l,r,c,   get=lambda: not self.input_file.nml_constraints.isDefault)
        Label       (l,r,c+1, get="Constraints", lab_disable=True)
        ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self._edit_constraints,
                     toolTip="Edit options")
        r += 1
        l.setRowStretch (r,2)
        # r += 1
        # Label       (l,r,c+1, get="Xoptfoil2 Input File", lab_disable=True)
        # ToolButton  (l,r,c+2, icon=Icon.EDIT, set=self._edit_input_file,
        #              toolTip="Direct editing of the Xoptfoil2 input file")
        
        l.setColumnMinimumWidth (0,20)
        l.setColumnMinimumWidth (1,130)
        l.setColumnStretch (1,2)

        return l
     

    def _edit_particle_swarm_options (self): 

        diag = Xo2_Particle_Swarm_Dialog (self, getter=self.input_file.nml_particle_swarm_options, 
                                            parentPos=(0.5, 0.0), dialogPos=(0.8,1.1))
        diag.exec()
        self.refresh ()
        self.app._on_xo2_input_changed()


    def _edit_xfoil_run_options (self): 

        diag = Xo2_Xfoil_Run_Dialog (self, getter=self.input_file.nml_xfoil_run_options, 
                                            parentPos=(0.5, 0.0), dialogPos=(0.8,1.1))
        diag.exec()
        self.refresh ()
        self.app._on_xo2_input_changed()


    def _edit_paneling_options (self): 

        diag = Xo2_Paneling_Dialog (self, getter=self.input_file.nml_paneling_options, 
                                            parentPos=(0.5, 0.0), dialogPos=(0.8,1.1))
        diag.exec()
        self.refresh ()
        self.app._on_xo2_input_changed()


    def _edit_constraints (self): 

        diag = Xo2_Constraints_Dialog (self, getter=self.input_file.nml_constraints, 
                                            parentPos=(0.5, 0.0), dialogPos=(0.8,1.1))
        diag.exec()
        self.refresh ()
        self.app._on_xo2_input_changed()


    def _edit_input_file (self):

        saved = self.input_file.save_nml ()

        if saved: 
            self.app._toast_message ("Options saved to Input file", toast_style=style.HINT)

        dialog = Xo2_Input_File_Dialog (self,self.input_file, parentPos=(0.8,0,9), dialogPos=(1,1))
        dialog.exec () 

        if dialog.result() == QDialog.DialogCode.Accepted:
            msg = "Input file successfully checked and saved"
            self.app._toast_message (msg, toast_style=style.GOOD)
            
            self.app._on_xo2_input_changed()



#!/usr/bin/env pythonbutton_color
# -*- coding: utf-8 -*-

"""  

Generic (compound) widgets based on original QWidgets

"""

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.WARNING)

import os
import types
from typing             import override
from enum               import Enum, StrEnum

from PyQt6.QtCore       import QSize, Qt, QMargins, pyqtSignal, QTimer, QEvent

from PyQt6.QtWidgets    import QLayout, QFormLayout, QGridLayout, QVBoxLayout, QHBoxLayout, QWIDGETSIZE_MAX
from PyQt6.QtWidgets    import (
                            QApplication, QWidget, QPushButton, QMenu,
                            QMainWindow, QLineEdit, QSpinBox, QDoubleSpinBox,
                            QLabel, QToolButton, QCheckBox,
                            QSpinBox, QComboBox, QSlider, QListWidget, QListWidgetItem,
                            QSizePolicy)
from PyQt6.QtGui        import QColor, QPalette, QFont, QIcon, QAction



#-------------------------------------------------------------------------------
# enums   
#-------------------------------------------------------------------------------


class button_style (Enum):
    """ button styles for Button widget"""

    PRIMARY         = 1                           # buttonstyle for highlighted button 
    SECONDARY       = 2                           # buttonstyle for normal action
    SUPTLE          = 3                           # buttonstyle for subtle appearance 
    ICON            = 4                           # buttonstyle for icon only button 
    RED             = 5                           # buttonstyle for red - stop - style 


class style (Enum):
    """ enums for style getter  - tuple for light and dark theme """
    # color see https://www.w3.org/TR/SVG11/types.html#ColorKeywords

                  #  light dark
    NORMAL        = (None, None)
    COMMENT       = ("dimgray","gray")  
    ERROR         = ('red', 'red')
    HINT          = ("dodgerblue", "dodgerblue")
    WARNING       = ('orange','orange')
    GOOD          = ('mediumspringgreen','mediumspringgreen')


class size (Enum):
    HEADER         = (13, QFont.Weight.Medium) # (13, QFont.Weight.ExtraLight)
    HEADER_SMALL   = (10, QFont.Weight.DemiBold)
    NORMAL         = ( 9, QFont.Weight.Normal) 
    SMALL          = ( 7, QFont.Weight.Normal) 

ALIGN_RIGHT         = Qt.AlignmentFlag.AlignRight
ALIGN_LEFT          = Qt.AlignmentFlag.AlignLeft
ALIGN_TOP           = Qt.AlignmentFlag.AlignTop
ALIGN_BOTTOM        = Qt.AlignmentFlag.AlignBottom




#-------------------------------------------------------------------------------
# Handle icons    
#-------------------------------------------------------------------------------

class Icon (QIcon):
    """ 
    Subclass of QIcon to handle named icons and provide an icon cache 

    Handle icon for light and dark mode.

    """
    
    # --- icons ---- 

    # <a target="_blank" href="https://icons8.com/icon/15813/pfeil%3A-einklappen">Pfeil: Einklappen</a> 
    # Icon von https://icons8.com
    # Windows 11 icon style 
    # color dark theme #C5C5C5, light theme #303030
    # size 96x96

    SETTINGS   = "settings" 
    COLLAPSE   = "collapse" 
    EXPAND     = "expand" 
    OPEN       = "open"     
    EDIT       = "edit"            # https://icons8.com/icon/set/edit/family-windows--static
    DELETE     = "delete"          # https://icons8.com/icon/set/delete/family-windows--static
    ADD        = "add"      
    NEXT       = "next"     
    PREVIOUS   = "previous" 
    FIT        = "fit" 
    RESETVIEW  = "resetView" 
    AE         = "AE"               # Airfoil Editor
    SHOW_INFO  = "info"

    # for messageBox 
    SUCCESS    = "success.png" 
    WARNING    = "warning.png" 
    INFO       = "info.png" 
    ERROR      = "error.png" 


    cache = {}

    subdirs = ["icons",             # py :  __file__: ...\modules\base    icons in: ...\base\icons      (toolbutton) 
               "..",                # py :  __file__: ...\modules\base    icons in: ...\modules         (app icon) 
               "..\\icons",         # exe:  __file__: ...\_internal\base\ icons in: ...\_internal\icons
               "."]

    @staticmethod
    def _get_icon(icon_name : str, light_mode = None, icon_dir=None) -> QIcon:
        """ load icon_name from file and store into class dict (cache) """

        ico = None

        if light_mode == True:
            icon_filename = icon_name + '_light' + '.png'
        elif light_mode == False:
            icon_filename = icon_name + '_dark' + '.png'
        else: 
            icon_filename = icon_name
        
        if Icon.cache.get (icon_filename, None) is not None:
            # icon in cache 
            ico =  Icon.cache.get (icon_filename)

        else: 

            # try to find in one of the subdirs relativ to __file__ (widgets.py) which can be in a subtree 
            root_dirs = [os.path.dirname(os.path.realpath(__file__))]

            # add optional icon_dir to root_dirs 
            if icon_dir:
                root_dirs.append(icon_dir)

            wrong_dirs = []

            for root_dir in root_dirs: 
                for subdir in Icon.subdirs:
                    icon_dir = os.path.join(root_dir, subdir)
                    icon_pathFilename = os.path.join(icon_dir, icon_filename)

                    # found icon file 
                    if os.path.isfile (icon_pathFilename):
                        ico = QIcon (icon_pathFilename)
                        Icon.cache [icon_filename] = ico 
                        break 
                    else: 
                        wrong_dirs.append(icon_dir)

        if ico is None:                 
            logger.error (f"Icon '{icon_filename} not found in {wrong_dirs}")

        return ico 


    @override
    def __init__ (self, aName, light_mode=None, icon_dir=None):
        """ Allow an Icon name to create a QIcon

        Args:
            aName:  name of icon
            light_mode: retrieve icon for light or dark mode
            icon_dir: optional new icon dir to look additionally for icon file 
        """

        if isinstance (aName, str): 
            super().__init__ (Icon._get_icon (aName, light_mode=light_mode, icon_dir=icon_dir))
        else:
            super().__init__(aName) 



#-------------------------------------------------------------------------------
# Helper functions   
#-------------------------------------------------------------------------------

def set_background (aWidget : QWidget, 
                     darker_factor : int | None = None,
                     color : QColor | int | None  = None,
                     alpha : float | None = None):
    """ 
    Set background color of a QWidget either by
        - darker_factor > 100  
        - color: QColor or string for new color
        - alpha: transparency 0..1 
    Returns the QPalette before changes were applied
    """

    new_color = None 
    aWidget.setAutoFillBackground(True)

    palette_org = aWidget.palette()
    palette     = aWidget.palette()

    if darker_factor:
        new_color = palette.color(QPalette.ColorRole.Window).darker (darker_factor)
    elif color:
        new_color = QColor(color) 
        if alpha: 
            new_color.setAlphaF (alpha)

    if new_color:
        palette.setColor(QPalette.ColorRole.Window, new_color)
        aWidget.setPalette(palette)

    return palette_org


#-------------------------------------------------------------------------------
# Widgets  
#-------------------------------------------------------------------------------

class Widget:
    """
    Extends QtWidgets to get an "access path" usage of widgets
        where a widget 'get' and 'set' its data by itself
        or by property with 'obj' and 'prop'

    A widgets add itsself to a 'QLayout' 

    Signals:

        sig_changed         a Value in an input field was changed and set into object 

    """


    # Signals

    sig_changed  = pyqtSignal(object)                   # (Object class name, Method as string, new value)

    # constants 

    LIGHT_INDEX = 0                                     # = Qt color index 
    DARK_INDEX  = 1 

    light_mode = True                                   # common setting of light/dark mode 
                                       
    _width  = None
    _height = None 


    def __init__(self,
                 layout: QLayout, 
                 *args,                                 # optional: row:int, col:int, 
                 rowSpan = 1, colSpan=1, 
                 align : Qt.AlignmentFlag = None,       # alignment within layout 
                 width :int = None,
                 height:int = None,
                 obj = None,                            # object | bound method  
                 prop = None,                           # Property 
                 get = None, 
                 set = None, 
                 signal : bool | None = None, 
                 id = None,
                 disable = None, 
                 hide = None,                           
                 style = style.NORMAL,                  # color style 
                 styleRole = QPalette.ColorRole.Base,   # default apply style to background
                 fontSize = size.NORMAL,
                 toolTip = None,
                 orientation = None): 
        
        # needed to build reference so self won't be garbage collected 
        if orientation is not None:         # special handling for slider which needs arg orientation
            super().__init__ (orientation)
        else: 
            super().__init__ ()

        self._layout = layout  

        self._row = None                    # optional positional arguments 
        self._col = None 
        if len(args) > 0: 
            self._row = args[0]
        if len(args) > 1: 
            self._col = args[1] 

        self._rowSpan = rowSpan
        self._colSpan = colSpan 
        self._alignment = align 
        self._width  = width  if width  is not None else self._width
        self._height = height if height is not None else self._height

        # "get & set" or "obj & prop" variant    

        self._val = None

        if obj is not None and prop is None: 
            raise ValueError (f"{self}: argment 'prop' is missing")
        if obj is None and prop is not None: 
            raise ValueError (f"{self}: argument 'obj' is missing")
        if prop is not None and get is not None: 
            raise ValueError (f"{self}: arguments 'obj' and 'get' can't be mixed together")

        if isinstance (prop, property) and obj is not None: 
            self._obj    = obj
            self._prop   = prop  
            self._getter = self._get_getter_of_property (obj, prop)  # function out of property                                    # property
            self._setter = self._get_setter_of_property (obj, prop)  # function out of property 
        else:
            self._obj    = None 
            self._prop   = None 
            self._getter = get                                      # bound method or None 
            self._setter = set 

        self._id = id 

        self._while_setting = False 

        # handle disable / hide  

        self._disable_in_refresh = False                           # temp overwrite in refresh 
        self._disabled_getter = None
        if isinstance(disable, bool):
            self._disabled   = disable                     
        elif callable (disable):
            self._disabled   = None                          
            self._disabled_getter = disable
        else: 
            self._disabled   = False                                # default values 

        if self._setter is None and disable is None and self._prop is None:
            self._disabled = True  

        self._hidden   = False                          
        if isinstance(hide, bool):
            self._hidden_getter = hide
        elif callable (hide):
            self._hidden_getter = hide
        else: 
            self._hidden_getter = None

        # tooltip

        self._toolTip = toolTip

        # style of widget 

        self._style_getter = style 
        self._style = None 
        self._style_role = styleRole                        # apply to background or text 
        self._font = fontSize

        self._palette_normal = self._initial_palette()      # will be copy of palette - for style reset  

        # emit signal 

        self._signal = signal if isinstance (signal, bool) else True  

        # connect to parent refresh signal 



    def __repr__(self) -> str:
        # overwritten to get a nice print string 
        text = f" '{str(self._val)}'" if self._val is not None else ''
        return f"<{type(self).__name__}{text} {id(self)}>"


    #--- static helper functions

    @staticmethod
    def refresh_childs (parent: QWidget):
        """ refresh all childs of parent"""
        w : Widget
        for w in parent.findChildren (Widget):
            w.refresh() 


    @staticmethod
    def _set_height (widget :QWidget , height):
        """ set self min/max height """
        if height is None: 
            return 
        elif isinstance (height, tuple):
            min_height = height[0]
            max_height = height[1]
        else:
            min_height = height
            max_height = height

        min_height = min_height if min_height else QWIDGETSIZE_MAX
        max_height = max_height if max_height else QWIDGETSIZE_MAX
        widget.setMinimumHeight(min_height)
        widget.setMaximumHeight(max_height)        


    @staticmethod
    def _set_width (widget :QWidget , width):
        """ set self min/max width """
        if width is None: 
            return 
        elif isinstance (width, tuple):
            min_width = width[0]
            max_width = width[1]
        else:
            min_width = width
            max_width = width
        min_width = min_width if min_width else QWIDGETSIZE_MAX
        max_width = max_width if max_width else QWIDGETSIZE_MAX
        widget.setMinimumWidth(min_width)
        widget.setMaximumWidth(max_width)


    #---  public methods 

    def refresh (self, disable : bool|None = False):
        """
        Refresh self by re-reading the 'getter' path 
            - disable: optional overwrite of widgets internal disable state  
        """

        # do not refresh if self will be now hidden

        self._hidden = self._get_value (self._hidden_getter, default=self._hidden)
        self._set_Qwidget_hidden ()

        if self._hidden:
            return 
        
        # avoid circular actions with refresh

        if self._while_setting:                          
            logger.debug (str(self) + " - refresh while set callback")

            #leave callback and refresh in a few ms 
            timer = QTimer()                                
            timer.singleShot(20, self.refresh)     # delayed emit 
        
        else: 

            # optionally overwrite disabled in refresh 

            if isinstance (disable, bool):
                self._disable_in_refresh = disable 

            # In case of using 'property' for get/set:
            #   the object could have been changed (e.g. subclass like Geometry_Bezier)
            if self._obj is not None and self._prop is not None:            
                self._getter = self._get_getter_of_property (self._obj, self._prop)  
                self._setter = self._get_setter_of_property (self._obj, self._prop)  

            self._get_properties ()

            # logger.debug (f"{self} - refresh (disable={disable} -> {self._disabled})")

            self._set_Qwidget (refresh=True)



    #---  from / to outside 

    def _get_properties (self): 
        """
        Read all the properties like disablee, style as they can be 
            - bound methods
            - fixed values 
            - property (only for self._val)
        """
        # should be overridden for additional properties 

        self._val       = self._get_value (self._getter, obj=self._obj, id= self._id)

        self._disabled  = self._get_value (self._disabled_getter, default=self._disabled)
        self._hidden    = self._get_value (self._hidden_getter, default=self._hidden)
        self._style     = self._get_value (self._style_getter)
        
 
    def _get_value(self, getter, id=None, obj=None, default=None):
        """
        Read the value. 'getter' can be 
            - bound method 
            - object of basic type: str, int, float as widget value 
        Optional 'id' is used as argument of bound method 'getter'
        'default' is taken, if 'getter' results in None 
        """
        try: 
            if isinstance (getter, property):               # getter is a property of obj 
                o = obj() if callable (obj) else obj
                if o is not None:
                    val = getter.__get__(o, type(o))
                else: 
                    val = None 

            elif callable(getter):                          # getter is a bound method ?
                if not id is None:                          # an object Id was set to identify object
                    val =  getter(id=self._id) 
                else:            
                    val =  getter()                         # normal callback
            else:                                           # ... no - getter is base type 
                val =  getter 
        except AttributeError:                              # access path of getter could be currently None
            val = None 

        if val is None and default is not None: 
            val = default

        return val 


    def _get_property_value(self, obj : object, obj_property : property):
        """
        Read a value of the 'obj_property' in model object 'obj'.
        """
        o = obj() if callable (obj) else obj            # obj either bound methed or object

        return  obj_property.__get__(o, type(o))        # access value of property in obj 



    def  _get_getter_of_property (self, obj , obj_property : property | str) : 
        """ 
        build a property function like 'thickness' out of the (get) property
        for the object 'obj.' (ensuring 'obj' has this property)
        'obj_property' can either be a class property or a string representing the property
        """

        o = obj() if callable (obj) else obj            # obj either bound methed or object

        if isinstance (obj_property, property):
            prop_name = obj_property.fget.__name__ 
        else: 
            prop_name = obj_property

        if hasattr (o.__class__, prop_name):
            getter = getattr (o.__class__, prop_name)
        else: 
            getter = None   
            logger.debug (f"{self} setter function '{prop_name} does not exist in {o.__class__}")

        return getter 



    def  _get_setter_of_property (self, obj , obj_property : property | str) : 
        """ 
        build a setter function like 'set_thickness(...)' out of the (get) property
        for the object 'obj.'
        'obj_property' can either be a class property or a string representing the property
        """

        o = obj() if callable (obj) else obj            # obj either bound methed or object

        if isinstance (obj_property, property):
            prop_name = obj_property.fget.__name__ 
        else: 
            prop_name = obj_property

        set_name  = "set_" + prop_name  
      
        if hasattr (o.__class__, set_name):
            setter = getattr (o.__class__, set_name)
        else: 
            setter = None   
            logger.debug (f"{self} setter function '{set_name} does not exist in {o.__class__}")

        return setter 



    def _set_value (self, newVal):
        """write the current value of the widget to model via setter path
        """

        if newVal is None:                          # None for button 
            pass
        elif self._val == newVal :                  # different signals could have beem emitted
            return
        else:                                        
            self._val = newVal                      # keep value for change detection

         # avoid circular actions with refresh()
        self._while_setting = True                 

        have_called = self._set_value_callback ()

        if have_called and self._signal: 
            self._emit_change () 

        self._while_setting = False                


    def _set_value_callback (self):
        """ 
        Do the set callback into the parent with new value 
            - return True if call was done 
        """

        have_called = False 

        # set value if setter is property and object 

        if self._obj is not None and isinstance (self._setter, types.FunctionType):            

            obj = self._obj() if callable (self._obj) else self._obj
            if self._val is None:                   # typically a button method which has no arg
                if self._id is None:                # an object Id was set to identify object
                    self._setter(obj)               # normal callback
                else:            
                    self._setter(obj, id=self._id) 
            else:                                   # a method like: def myMth(self, newVal)
                if self._id is None:                # an id was set to identify object?
                    self._setter(obj, self._val)    # normal callback
                else:            
                    self._setter(obj, self._val, id=self._id) 

            have_called = True

        # set value is setter is bound method or function (lambda)

        elif callable(self._setter):                # setter is a method ?
            if self._val is None:                   # typically a button method which has no arg
                if self._id is None:                # an object Id was set to identify object
                    self._setter()                  # normal callback
                else:            
                    self._setter(id=self._id) 
            else:                                   # a method like: def myMth(self, newVal)
                if self._id is None:                # an id was set to identify object?
                    self._setter(self._val)         # normal callback
                else:            
                    self._setter(self._val, id=self._id) 

            have_called = True

        return have_called


    def _emit_change (self):
        """ emit change signal""" 

        if isinstance(self._setter, types.FunctionType):
            qualname  = self._setter.__qualname__
        elif callable(self._setter):        
            qualname  = self._setter.__qualname__
        else: 
            qualname = ''

        logger.debug (f"{self} emit sig_changed: {qualname} ({self._val})")
        self.sig_changed.emit (self)
        # emit signal delayed so we leave the scope of Widget 
        # timer = QTimer()                                
        # timer.singleShot(50, self.sig_changed.emit)     # delayed emit 


    def _layout_add (self, widget = None, col = None):
        """ adds self to layout"""

        if widget is None: 
            widget : QWidget = self

        if col is None:                         # alternate col 
            col = self._col

        if isinstance (self._layout, QGridLayout): 
            if self._alignment is None: 
                self._layout.addWidget(widget, self._row, col, 
                                                self._rowSpan, self._colSpan)
            else: 
                self._layout.addWidget(widget, self._row, col, 
                                                self._rowSpan, self._colSpan,
                                                alignment = self._alignment)
        elif isinstance (self._layout, (QFormLayout,QVBoxLayout, QHBoxLayout)):
            self._layout.addWidget(widget)

        # strange (bug?): if layout is on main window the stretching works as expected
        # on a sub layout the widget doesn't stretch if on widget in the column is fixed 
        widget.setSizePolicy( QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed )


    #---  from / to QWidget - inside the widget 

    def _should_be_disabled (self) -> bool:
        """ True if self currently should be disabled but maybe not set into QWidget"""
        return self._disabled or self._disable_in_refresh
    

    def _set_Qwidget (self, refresh=False): 
        """ set value and properties of self Qwidget"""
        # must be overridden 

        # minimum for all 
        self._set_Qwidget_hidden ()
        self._set_Qwidget_disabled ()
        self._set_Qwidget_style ()              # NORMAL, WARNING, etc 
        self._set_QWidget_toolTip ()


    def _set_Qwidget_static (self, widget = None): 
        """ set static properties of self Qwidget like width"""
        # can be overlaoed 

        if widget is None: 
            widget : QWidget = self

        # set font 
        if self._font != size.NORMAL:
            font = widget.font() 
            font.setPointSize(self._font.value[0])
            font.setWeight   (self._font.value[1])  
            widget.setFont(font)

        # set width and height 
        Widget._set_width  (widget, self._width)
        Widget._set_height (widget, self._height)


    def _set_QWidget_toolTip (self):
        """ set a toolTip into self QWidget"""

        if self._toolTip is not None:
            toolTip = self._toolTip() if callable (self._toolTip) else self._toolTip
            widget : QWidget = self
            widget.setToolTip (toolTip)


    def _set_Qwidget_disabled (self):
        """ set self Qwidget according to self._disabled"""
        # can be overridden to suppress enable/disable 

        widget : QWidget = self
        if widget.isEnabled() != (not self._should_be_disabled()) :
            widget.setDisabled(self._should_be_disabled())


    def _set_Qwidget_hidden (self):
        """ set self Qwidget according to self._hidden"""

        # can be overridden to suppress hiden 

        widget : QWidget = self

        if widget.parentWidget() is None:
            # if self still doesn't have a parent, do not setVisible(True) as this would lead to a ghost window
            #   but set not visible in any case 
            if self._hidden:
                widget.setVisible(False)
                
        elif widget.isHidden() != self._hidden :            # use hidden as self could be not visible because of parent
                widget.setVisible(not self._hidden)


    def _set_Qwidget_style (self): 
        """ set (color) style of QWidget based on self._style"""

        # default color_role is change the text color according to style

        # QPalette.ColorRole.Text, .ColorRole.WindowText, .ColorRole.Base, .ColorRole.Window
        self._set_Qwidget_style_color (self._style, self._style_role)  


    def _set_Qwidget_style_color (self, aStyle : style, color_role : QPalette.ColorRole= None):
        """ 
        low level set of colored part of widget accordings to style and color_role
            color_role = .Text, .Base (background color), .Window
        """

        # # qt so new backgrund colour will be applied  - but: no more transparent inheritence!
        autoFill = False                                        

        if aStyle in [style.WARNING, style.ERROR, style.COMMENT, style.GOOD, style.HINT]:

            index = self.LIGHT_INDEX if self.light_mode else self.DARK_INDEX

            color          = QColor (aStyle.value[index])
            color_disabled = QColor (color)

            # if it's background color apply alpha
            if color_role in [QPalette.ColorRole.Base, QPalette.ColorRole.Window, QPalette.ColorRole.Button]:
                if aStyle == style.GOOD or aStyle == style.HINT:
                    color.setAlphaF (0.3) 
                    color_disabled.setAlphaF (0.15)
                else:
                    color.setAlphaF (0.15)
                    color_disabled.setAlphaF (0.1)
                autoFill = True                                         # apply background 
            elif color_role in [QPalette.ColorRole.WindowText, QPalette.ColorRole.Text]:
                pass
            else: 
                raise ValueError (f"ColorRole {color_role} not supported")

            if self._palette_normal is None: 
                palette =  self._initial_palette ()
                self._palette_normal = palette if palette else self.palette()

            palette =  QPalette (self._palette_normal)          # copy normal palette 
            palette.setColor(QPalette.ColorGroup.Active,   color_role, color)
            palette.setColor(QPalette.ColorGroup.Inactive, color_role, color)
            palette.setColor(QPalette.ColorGroup.Disabled, color_role, color_disabled)

            self._update_palette (palette, autoFill=autoFill)                     

        elif aStyle == style.NORMAL :

            self._reset_palette ()


    def _initial_palette (self) -> QPalette:
        """ returns initial normal palette of self"""

        return None 


    def _update_palette (self, palette: QPalette, qwidget : QWidget = None, autoFill=False):
        """ set new QPalette for self or qwidget if set """

        if self._palette_normal is None: return 
    
        # to overwrite if self is compound like QSpinbox ()  
        if qwidget is None: 
            qwidget : QWidget = self

        if qwidget.palette() != palette:
            qwidget.setAutoFillBackground(autoFill)             # important to be applied
            qwidget.setPalette (palette) 


    def _reset_palette (self):
        """ reset self (or of qwidget) palette to palette normal """

        self._update_palette (self._palette_normal)


    def _getFrom_Qwidget (self):
        """returns the current value of QWidget  
        """
        # must be over written by specific widget 
        return None
    
                    
    @property     
    def _name (self): 
        """ name for error messages etc. """

        return self.__class__



# ----------------------------------------------------------------
#  abstract widget super classes 
# ----------------------------------------------------------------
 
class Field_With_Label (Widget):
    """
    Abstract Field which adds an optional Label before the widget 

    If 'lab':  
        col=i   : Label
        col=i+1 : Field | ComboBox 

        In this case, only QGridLayout is supported    
    """

    _height = 26 

    def __repr__(self) -> str:
        # overwritten to get a nice print string 
        lab = self._label._val if self._label else ""
        text = f" '{str(self._val)}'" if self._val is not None else ''
        return f"<{type(self).__name__} {lab}{text} {id(self)}>"
    


    def __init__(self, *args, 
                 lab : str | None = None,                   # label of field 
                 disable : bool | None = None,
                 lab_disable : bool = False,                # label will be disabled with field
                 toolTip : str|None = None,
                 **kwargs):

        self._label = None 

        super().__init__(*args, disable=disable, toolTip=toolTip, **kwargs)


        if lab is not None: 
            if not isinstance (self._layout, QGridLayout):
                raise ValueError (f"{self} Only QGridLayout supported for Field with a Label")
            else: 
                self._label = Label(self._layout, self._row, self._col, get=lab, 
                                    disable=disable, lab_disable=lab_disable, align=self._alignment,
                                    colSpan=1,              # label is always only 1 column
                                    rowSpan=self._rowSpan,  # rowSpan same as parent field
                                    toolTip=None)           # no tooltip on label 
            

    def _layout_add (self, widget=None):
        # overloaded

        # put into grid / layout - if there is a label one col more 
        if self._label:
            col = self._col + 1 if self._col is not None else None 
        else: 
            col = self._col 

        # strange (bug?): if layout is on main window the stretching works as expected
        # on a sub layoutthe widget doesn't stretch if on widget in the column is fixed 
        super()._layout_add (col=col, widget=widget)

        # if self._label and self._alignment is not None:
        #     self._layout.setAlignment (self._label, self._alignment)

        self.setSizePolicy( QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed )


    @override
    def _set_Qwidget (self, refresh=False): 
        """ set value and properties of self Qwidget"""
        super()._set_Qwidget (refresh=refresh)

        # overridden to also hide / disable self label 
        if self._label:
            self._label._set_Qwidget (refresh=refresh)


    @override
    def _set_Qwidget_hidden (self):
        """ set self Qwidget according to self._hidden"""

        super()._set_Qwidget_hidden ()

        # overridden to also hide / disable self label 
        if self._label:
            self._label._hidden = self._hidden
            self._label._set_Qwidget_hidden ()


    def _on_finished(self):
      self._set_value (self.text())



# ----------------------------------------------------------------
# -----------  real generic widget subclasses --------------------
# ----------------------------------------------------------------


class Label (Widget, QLabel):
    """
    label text with word wrap 
    """

    _height = 26 

    def __init__(self, *args, 
                 styleRole = QPalette.ColorRole.WindowText,  # for background: QPalette.ColorRole.Window
                 disable = None,                             
                 lab_disable : bool = False,                 # label can be disabled (color)
                 wordWrap = False,                           # activate word wrap
                 **kwargs):

        # special disable handling for Label as typically it is not disabled 
        self._label_disable = lab_disable

        # labels are typically not disabled 
        disable = False if disable is None else disable 

        super().__init__(*args, styleRole=styleRole, disable=disable,**kwargs)

        self._get_properties ()

        self._layout_add ()                                 # put into layout - so it gets parent early

        self._set_Qwidget_static ()
        if wordWrap:
            self.setWordWrap (wordWrap)

        self._set_Qwidget ()


       # self.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse)


    @override
    def _initial_palette(self) -> QPalette | None:
        """ returns initial normal palette of self"""

        palette =  self.palette()
        if not self.light_mode:                             # dark mode: make labels a little darker (not white)   
            palette =  self.palette()
            color = palette.color(QPalette.ColorRole.WindowText).darker (150)
            palette.setColor(QPalette.ColorRole.WindowText,color)
        return palette


    def _set_Qwidget (self, **kwargs):
        """ set value and properties of self Qwidget"""

        super()._set_Qwidget (**kwargs)
        self.setText (self._val)


    @override
    def _set_Qwidget_disabled (self):
        """ set self Qwidget according to self._disabled"""
        if self._label_disable:
            super()._set_Qwidget_disabled()
        else: 
            # do not disable Label (color ...) 
            pass


 
class Field (Field_With_Label, QLineEdit):
    """
    String entry field  
    """

    _height = 24 

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._get_properties ()

        # put into grid / layout 
        self._layout_add ()

        self._set_Qwidget_static ()
        self._set_Qwidget ()

        # connect signals 
        self.editingFinished.connect(self._on_finished)
        self.returnPressed.connect(self._on_finished)
        self.textEdited.connect(self._on_finished)              # also set with every edit 


    def _set_Qwidget (self, refresh=False, **kwargs):
        """ set value and properties of self Qwidget"""
        super()._set_Qwidget (refresh=refresh,**kwargs)
        val = self._val if self._val is not None else ''

        if refresh:
            # setText resets cursor position - save and restore in case of refresh 
            cursor_pos = self.cursorPosition() 
            self.setText (str(val))
            self.setCursorPosition (cursor_pos)
        else: 
            self.setText (str(val))


    @override
    def _initial_palette(self):
        """ returns initial normal palette of self"""

        palette =  self.palette()

        color = palette.color (QPalette.ColorGroup.Disabled, QPalette.ColorRole.Base)
        color.setAlphaF (0.3)
        palette.setColor (QPalette.ColorGroup.Disabled, QPalette.ColorRole.Base, color)

        if self.light_mode:
            color = palette.color (QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text)
            color = color.darker (180)
            palette.setColor (QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text, color)

        return palette


    def _on_finished(self):

      self._set_value (self.text())



class FieldI (Field_With_Label, QSpinBox):
    """
    Integer entry field with spin buttons
    """
    
    def __init__(self, *args, 
                 step = None,
                 lim = None, 
                 unit : str = None, 
                 specialText : str = None,   # Qt specialValueText
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._step = int(step) if step is not None else None
        self._spin = step is not None 
        self._unit = unit
        self._lim = None 
        self._lim_getter = lim
        self._specialText = specialText

        self._get_properties ()

        self._layout_add ()                                 # put into layout - so it gets parent early

        self._set_Qwidget_static ()
        self._set_Qwidget ()

        # connect signals 
        self.editingFinished.connect (self._on_finished)


    def _get_properties (self): 
        # overloaded
        super()._get_properties () 
        self._val = int (self._val) if self._val is not None else 0 
        self._lim = self._get_value (self._lim_getter)


    def _set_Qwidget (self, **kwargs):
        """ set value and properties of self Qwidget"""

        super()._set_Qwidget (**kwargs)

        if self._lim: 
            self.setRange (self._lim[0], self._lim[1])
        self.setValue (self._val)


    def _set_Qwidget_disabled (self):
        """ set self Qwidget according to self._disabled"""
        super()._set_Qwidget_disabled()

        # overloaded to show/hide spin buttons  
        # parent could be disabled - so also remove spin buttons 
        if self._spin and not self._should_be_disabled() and self.isEnabled():
            self.setButtonSymbols(QSpinBox.ButtonSymbols.PlusMinus)
        else: 
            self.setButtonSymbols(QSpinBox.ButtonSymbols.NoButtons)


    def _set_Qwidget_static (self): 
        """ set static properties of self Qwidget like width"""
        super()._set_Qwidget_static ()

        self.setRange (-999999999, 999999999)                  # Qt default is 0.0 and 99

        if self._unit: 
            self.setSuffix (" " + self._unit)
        if self._spin:
            self.setSingleStep (self._step)
            if self._specialText:
                self.setSpecialValueText (self._specialText)
        self.setAlignment(Qt.AlignmentFlag.AlignRight)

    @override
    def _initial_palette(self):
        """ returns initial normal palette of self"""

        palette =  self.palette()

        if self.light_mode:
            color = palette.color (QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text)
            color = color.darker (180)
            palette.setColor (QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text, color)

        return palette


    @override
    def _update_palette (self, palette: QPalette, autoFill=False):
        """ set new QPalette for self"""
        # for QSpinbox LineEdit is the relevant qwidget 
        # super()._update_palette (palette)
        super()._update_palette (palette, qwidget=self.lineEdit(), autoFill=autoFill) 


    def stepBy (self, step):
        # Qt overloaded: Detect when a Spin Button is pressed with new value 
        value = self.value()
        super().stepBy(step)
        if self.value() != value:
            self._on_finished ()


    def _on_finished(self):
      self._set_value (self.value())




class FieldF (Field_With_Label, QDoubleSpinBox):
    """
    Float entry field with spin buttons
    """

    _height = 26 

    def __init__(self, *args, 
                 step : float = None,
                 lim = None, 
                 unit : str = None,
                 dec : int = 2,
                 specialText : str = None,   # Qt specialValueText
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._step = float(step) if step is not None else None
        self._spin = step is not None 
        self._unit = unit
        self._dec  = int(dec) 
        self._lim = None 
        self._lim_getter = lim
        self._specialText = specialText

        self._get_properties ()

        self._layout_add ()                                 # put into layout - so it gets parent early

        self._set_Qwidget_static ()
        self.ensurePolished()
        self._set_Qwidget ()

        # a (new) style isn't apllied when done during __init__ - so do it later 
        if self._palette_normal is not None:            # indicates that a palette was set
            timer = QTimer()                                
            timer.singleShot(10, self._set_Qwidget_style)     

        # connect signals 
        self.editingFinished.connect(self._on_finished)


    @override
    def _get_properties (self): 
        super()._get_properties () 
        self._lim = self._get_value (self._lim_getter)


    def _set_Qwidget (self, **kwargs):
        """ set value and properties of self Qwidget"""

        super()._set_Qwidget (**kwargs)

        if self._lim: 
            self.setRange (self._lim[0], self._lim[1])

        # set val into Qwidget handle - percent unit automatically 
        if self._val is not None:
            if self._unit == '%':
                self.setValue (self._val * 100.0)
            else: 
                self.setValue (self._val)


    def _set_Qwidget_disabled (self):
        """ set self Qwidget according to self._disabled"""
        super()._set_Qwidget_disabled()

        # overloaded to show/hide spin buttons  
        # parent could be disabled - so also remove spin buttons 
        if self._spin and not self._should_be_disabled() and self.isEnabled():
            self.setButtonSymbols(QSpinBox.ButtonSymbols.PlusMinus)
        else: 
            self.setButtonSymbols(QSpinBox.ButtonSymbols.NoButtons)


    def _set_Qwidget_static (self): 
        """ set static properties of self Qwidget like width"""
        super()._set_Qwidget_static ()

        self.setRange (-999999999, 999999999)                  # Qt default is 0.0 and 99
        self.setDecimals(self._dec) 
        if self._unit: 
            self.setSuffix (" " + self._unit)
        if self._spin:
            self.setSingleStep (self._step)
            if self._specialText:
                self.setSpecialValueText (self._specialText)
        self.setAlignment(Qt.AlignmentFlag.AlignRight)


    @override
    def _initial_palette(self):
        """ returns initial normal palette of self"""

        palette =  self.palette()

        if self.light_mode:
            color = palette.color (QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text)
            color = color.darker (180)
            palette.setColor (QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text, color)

        return palette


    @override
    def _update_palette (self, palette: QPalette, autoFill=False):
        """ set new QPalette for self"""

        # for QDoubleSpinbox LineEdit is the relevant qwidget 
        # super()._update_palette (palette)
        super()._update_palette (palette, qwidget=self.lineEdit(), autoFill=autoFill) 


    def stepBy (self, step):
        # Qt overloaded: Detect when a Spin Button is pressed with new value 
        value = self.value()
        super().stepBy(step)
        if self.value() != value:
            self._on_finished ()


    def _on_finished(self):
        """ signal slot finished"""
        new_val = round(self.value(), self._dec)  # Qt sometimes has float artefacts 

        # check i f the value in percent changed based on decimals 
        if self._unit == '%':
            my_val  = round(self._val * 100, self._dec)
            if new_val == my_val: 
                return
            new_abs_val = round(new_val / 100.0, 8)
        else: 
            new_abs_val = new_val 
        # get val from Qwidget - handle percent unit automatically 

        self._set_value (new_abs_val)


class Slider (Widget, QSlider): 
    """
    
    A value slider - extends the QSlider to handle float values

    Default are 100 steps with a step size of 1% between the lim () 
    """

    _height = 26 

    def __init__(self, *args, 
                 step : float = None,
                 dec : int = 2,
                 lim = None, 
                 **kwargs):
        super().__init__(*args, orientation = Qt.Orientation.Horizontal,**kwargs)

        self._step = float(step) if step is not None else None
        self._dec  = int(dec) 
        self._lim = None 
        self._lim_getter = lim

        self._slider_val = None
        self._slider_min = 0 
        self._slider_max = 100 

        self._get_properties ()

        # if value is an Integer (dec=0) take lim as sliders internal min, max
        # to achieve a step size of "1"
        if dec==0 and self._lim:
            self._slider_min = self._lim[0] 
            self._slider_max = self._lim[1]    
        else:
            self._slider_min = 0                    # will be translated fore and back 
            self._slider_max = 100             

        self._layout_add ()                                 # put into layout - so it gets parent early

        self._set_Qwidget_static ()
        self._set_Qwidget ()

        # connect signals 
        # self.valueChanged.connect(self._on_changed)  - would be also set, if value is set into slider
        self.sliderMoved.connect(self._on_changed)


    @override
    def _get_properties (self): 
        super()._get_properties () 
        self._lim = self._get_value (self._lim_getter)


    @override
    def _set_Qwidget_static (self): 
        """ set static properties of self Qwidget like width"""
        super()._set_Qwidget_static ()

        self.setMinimum (self._slider_min)
        self.setMaximum (self._slider_max)
        self.setSingleStep (1) 


    @override
    def _set_Qwidget (self, **kwargs):
        """ set value and properties of self Qwidget"""
        super()._set_Qwidget (**kwargs)

        self._slider_val = self._from_val_to_slider ()
        self.setValue (self._slider_val)


    def _on_changed(self, slider_val):
        """ slot - finished slider movement"""

        self._slider_val = slider_val
        newVal = self._from_slider_to_val ()
        newVal = round(newVal, self._dec)                   # Qt sometimes has float artefacts 
        self._set_value (newVal)


    def _from_val_to_slider (self) -> int:
        """ get the slider value from _val"""

        if not isinstance (self._lim, tuple): 
            raise ValueError (f"{self} slider limits are not set")

        min_val = self._lim[0]
        max_val = self._lim[1]

        self._val = max (min_val, self._val)
        self._val = min (max_val, self._val)
        rel_val   = (self._val - min_val) / (max_val - min_val)

        min_slider = self._slider_min
        max_slider = self._slider_max

        slider_val = min_slider + rel_val * (max_slider - min_slider)
        return int (round(slider_val,0))


    def _from_slider_to_val (self):
        """ get widgets value from slider value (which is int)"""

        min_val = self._lim[0]
        max_val = self._lim[1]

        s = (self._slider_val - self._slider_min) / (self._slider_max - self._slider_min)

        return min_val + s * (max_val - min_val) 




class Button (Widget, QPushButton):
    """
    Button 
        - gets its label via 'text' 
        - when clicked, 'set' is called without argument 
    """
    def __init__(self, *args,
                 signal = False,                                # default is not to signal change 
                 text = None, 
                 button_style = button_style.SECONDARY,
                 styleRole = QPalette.ColorRole.Button,         # button has different color role 
                 **kwargs):
        super().__init__(*args, signal=signal, styleRole=styleRole, **kwargs)

        self._text = text 
        self._button_style = None 
        self._button_style_getter = button_style

        self._get_properties ()
        self._layout_add ()                                 # put into layout - so it gets parent early
        self._set_Qwidget_static ()
        self._set_Qwidget ()

        # connect signals 
        self.pressed.connect(lambda: self._set_value(None))


    def __repr__(self) -> str:
        text = f" '{str(self._text)}'" if self._text is not None else ''
        return f"<{type(self).__name__}{text}>"

    @override
    def _initial_palette(self) -> QPalette | None:
        """ returns initial normal palette of self"""

        palette =  self.palette()
        if not self.light_mode:                             # dark mode: make labels a little darker (not white)   
            palette =  self.palette()
            color = palette.color(QPalette.ColorRole.ButtonText).darker (120)
            palette.setColor(QPalette.ColorRole.ButtonText,color)
        return palette


    @override
    def _get_properties (self): 
        " get all the properties like disable"
        super()._get_properties () 
        self._button_style = self._get_value (self._button_style_getter)


    @override
    def _set_Qwidget_static (self): 
        """ set static properties of self Qwidget like width"""
        super()._set_Qwidget_static ()
        self.setText (self._text)


    @override
    def _set_Qwidget (self, **kwargs): 
        """ set properties of self Qwidget like data"""
        super()._set_Qwidget (**kwargs)        
        if self._button_style == button_style.PRIMARY:
            self.setDefault (True)
        else: 
            self.setDefault(False)



class ToolButton (Widget, QToolButton):
    """
    Icon tool button 
        - when clicked, 'set' is called without argument 
    """

    _width  = 22
    _height = 22 

    def __init__(self, *args, 
                 icon : str =None, 
                 signal = False,            # default is not to signal change 
                 **kwargs):
        super().__init__(*args, signal=signal,**kwargs)

        self._icon_name = icon                           # icon name 

        self._get_properties ()
        self._layout_add ()                                 # put into layout - so it gets parent early
        self._set_Qwidget_static ()
        self._set_Qwidget ()

        self.clicked.connect(lambda: self._set_value(None))


    @override
    def __repr__(self) -> str:
        # get a nice print string 
        text = f" '{str(self._icon_name)}'" if self._icon_name is not None else ''
        return f"<{type(self).__name__}{text}>"


    @override
    def _set_Qwidget_static (self): 
        """ set static properties of self Qwidget like width"""
        super()._set_Qwidget_static ()

        self.setAutoRaise (True)            # button frame only on focus

        if self._icon_name is not None: 

            icon_qt = Icon (self._icon_name, self.light_mode)
            if icon_qt is not None: 
                self.setIcon (icon_qt)
                self.setIconSize (QSize(16,16))
            else: 
                pass


class MenuButton (Button, QPushButton):
    """
    Push-Button with an action menu. 
    
    A QMenu (having actions) has to be provided instead of a 'set' method.

    """
    def __init__(self, *args,
                 menu : QMenu = None,                                # default is not to signal change 
                 **kwargs):
        
        self._menu = menu

        super().__init__(*args, **kwargs)

        if isinstance(menu, QMenu):
            self.setMenu (menu)
            self._disabled = None                                    # missing setter would disable
            self._set_Qwidget_disabled ()
        else: 
            self._disabled = True 


    @override
    def refresh(self, disable = False):
        """ overridden to also refresh MenuAction"""
        super().refresh(disable)
    
        for menuAction in self._menu.actions():
            if isinstance (menuAction, MenuAction):
                menuAction.refresh()


class MenuAction (QAction):
    """
    Menu item - QAction subclass to make it compatible with WIdget  
    """
    def __init__(self, 
                 text,                                              # menu text
                 parent,                                            # strange: QAction (still) needs a dummy parent
                 set = None,                                        # slot to be triggered 
                 disable = None,
                 toolTip = None,
                 **kwargs):
        super().__init__(text, parent, **kwargs)

        # conect self triggered to 'set' 
         
        if callable (set):
            self.triggered.connect(set)
        else:
            raise ValueError(f"{self} set method isn't callable")

        # init disable 

        self._disabled_getter = None
        if isinstance(disable, bool):
            self._disabled   = disable                     
        elif callable (disable):
            self._disabled   = None                          
            self._disabled_getter = disable
        else: 
            self._disabled   = False                                # default values 

        self._disabled  = self._get_value (self._disabled_getter, default=self._disabled)
        self.setDisabled (self._disabled)

        # tooltip 

        if toolTip: 
            self.setToolTip (toolTip)



    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.text()}>"


    def refresh (self):
        """ Refresh self disable state """

        self._disabled  = self._get_value (self._disabled_getter, default=self._disabled)
        self.setDisabled (self._disabled)


    def _get_value(self, getter, default=None):
        """ Read the value. 'getter' shall be bound method. 'default' is taken, if 'getter' results in None  """

        try: 
            if callable(getter):                            # getter is a bound method ?
                val =  getter()                             # normal callback
            else:                                           # ... no - getter is base type 
                val =  getter 
        except AttributeError:                              # access path of getter could be currently None
            val = None 

        if val is None and default is not None: 
            val = default

        return val 



class CheckBox (Widget, QCheckBox):
    """
    Checkbox 
        - gets its label via 'text' which is either string or getter 
        - when clicked, 'set' is called with argument of checkSTate 
    """

    _height = 26 

    def __init__(self, *args, 
                 text=None,
                 styleRole = QPalette.ColorRole.WindowText,
                 **kwargs):
        super().__init__(*args, styleRole=styleRole, **kwargs)

        self._text = None 
        self._text_getter = text 

        self._get_properties ()

        # put into grid / layout 
        self._layout_add ()

        self._set_Qwidget_static ()
        self._set_Qwidget ()

        # connect signals 
        self.checkStateChanged.connect(self._on_checked)


    def _get_properties (self): 
        """ get properties from parent"""
        super()._get_properties () 
        self._val = self._val is True 
        self._text = self._get_value (self._text_getter, default='')


    def _set_Qwidget (self, **kwargs):
        """ set value and properties of self Qwidget"""
        super()._set_Qwidget (**kwargs)
        self.setChecked (self._val)
        self.setText (self._text)


    def _set_Qwidget_static (self): 
        """ set static properties of self Qwidget like width"""
        super()._set_Qwidget_static ()


    @override
    def _initial_palette(self):
        """ returns initial normal palette of self"""

        palette =  self.palette()

        if not self.light_mode:
            palette =  self.palette()
            color = palette.color(QPalette.ColorRole.WindowText).darker (140)
            palette.setColor(QPalette.ColorRole.WindowText,color)

            color = palette.color (QPalette.ColorGroup.Disabled, QPalette.ColorRole.WindowText)
            color = color.darker (180)
            palette.setColor (QPalette.ColorGroup.Disabled, QPalette.ColorRole.WindowText, color)

        return palette


    def _on_checked(self):
      self._set_value (self.isChecked())



class ComboBox (Field_With_Label, QComboBox):
    """
    ComboBox  
        - values list via 'options' (bound method or list)
        - when clicked, 'set' is called argument with selected text as argument 
    """
        
    _height = 24 

    def __init__(self, *args, 
                 options = [], 
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._options_getter = options
        self._options = None

        self._get_properties ()

        # put into grid / layout 
        self._layout_add ()

        self._set_Qwidget_static ()
        self._set_Qwidget ()

        # connect signals 
        self.activated.connect(self._on_selected)


    @override
    def _initial_palette(self) -> QPalette | None:
        """ returns initial normal palette of self"""

        palette =  self.palette()
        if not self.light_mode:                             # dark mode: make text a little darker (not white)   
            palette =  self.palette()
            color = palette.color(QPalette.ColorRole.ButtonText).darker (120)
            palette.setColor(QPalette.ColorRole.ButtonText,color)
        return palette


    def _get_properties (self): 
        # overloaded
        super()._get_properties () 
        self._options = self._get_value (self._options_getter)
        self._val = str(self._val) if self._val is not None else None


    def _set_Qwidget (self, **kwargs):
        """ set value and properties of self Qwidget"""

        super()._set_Qwidget (**kwargs)

        self.clear()                                # addItems append to list
        self.addItems (self._options)
        self.setCurrentText (self._val)

        self._set_placeholder_text()


    def _on_selected (self):
      self._set_value (self.currentText())
      self._set_placeholder_text()


    def _set_placeholder_text (self):
        if not self._val and len(self._options) > 1: 
            self.setPlaceholderText ("Select")
            self.setCurrentIndex (-1)



class ComboSpinBox (Field_With_Label, QComboBox):
    """
    ComboBox  with spin buttons 'next' and 'rev'
        - values list via 'options' (bound method or list)
        - when clicked, 'set' is called argument with selected text as argument 
    """
    def __init__(self, *args, 
                 options = [], 
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._options_getter = options
        self._options = None
        self._get_properties ()

        if self._toolTip is None: 
            self._toolTip = 'Use mouse wheel to browse through items'

        # add spin buttons on helper widget 

        helper   = QWidget()
        l_helper = QHBoxLayout ()
        l_helper.setSpacing (1)
        l_helper.setContentsMargins (QMargins(0, 0, 0, 0))
        l_helper.addWidget (self,stretch=2) 

        self._wButton_prev = ToolButton (l_helper, icon=Icon.PREVIOUS,  set=self._on_pressed_prev, disable=self._prev_disabled)
        self._wButton_next = ToolButton (l_helper, icon=Icon.NEXT, set=self._on_pressed_next, disable=self._next_disabled)
        helper.setLayout (l_helper) 

        self._wButton_prev.setAutoRepeat(True)
        self._wButton_next.setAutoRepeat(True)

        self._set_Qwidget_static (widget=helper)
        self._set_Qwidget ()

        # refresh prev/next buttons to set now actual disable state 
        self._refresh_buttons()

        # put into grid / layout 
        self._layout_add (widget=helper)

        # connect signals 
        self.activated.connect(self._on_selected)
        # self.currentTextChanged.connect(self._on_selected)


    @override
    def _initial_palette(self) -> QPalette | None:
        """ returns initial normal palette of self"""

        palette =  self.palette()
        if not self.light_mode:                             # dark mode: make text a little darker (not white)   
            palette =  self.palette()
            color = palette.color(QPalette.ColorRole.ButtonText).darker (120)
            palette.setColor(QPalette.ColorRole.ButtonText,color)
        return palette


    @override
    def refresh (self, disable : bool|None = None): 
        """ refresh self """
        #overloaded to refresh also (state) of spin buttons
        super().refresh(disable=disable) 
        self._refresh_buttons () 


    def _refresh_buttons (self):
        """ refresh (disable) spin buttons """
        self._wButton_prev._get_properties ()
        self._wButton_prev._set_Qwidget ()
        self._wButton_next._get_properties ()
        self._wButton_next._set_Qwidget ()


    def _on_pressed_prev (self): 
        """ prev button pressed """
        cur_index = self.currentIndex() 
        if cur_index > 0:
            self.setCurrentIndex (cur_index - 1)                        # selection index of combobox
            self._refresh_buttons ()
            self._on_selected ()                                        # handle callback


    def _on_pressed_next (self): 
        " next button pressed  """
        cur_index = self.currentIndex() 
        if cur_index < (len(self._options)) - 1:
            self.setCurrentIndex (cur_index + 1)                        # selection index of combobox
            self._refresh_buttons ()
            self._on_selected ()                                        # handle callback


    def _prev_disabled (self) -> bool: 
        """ prev button disabled"""
        return self.currentIndex() <= 0


    def _next_disabled (self) -> bool: 
        """ prev button disabled"""
        return self.currentIndex() >= (len(self._options)) - 1


    def _get_properties (self): 
        # overloaded
        super()._get_properties () 
        self._options = self._get_value (self._options_getter)
        self._val = str(self._val)


    def _set_Qwidget (self, **kwargs):
        """ set value and properties of self Qwidget"""

        super()._set_Qwidget (**kwargs)

        self.clear()                                # addItems append to list
        self.addItems (self._options)
        self.setCurrentText (self._val)


    def _on_selected (self):
      """ slot - comboxbox value choosen """
      self._refresh_buttons ()
      self._set_value (self.currentText())


class ListBox (Field_With_Label, QListWidget):
    """
    Listbox  
        - values list via 'options' (method or list)
        - when clicked, 'set' is called argument with selected text as argument 
        - when double clicked, in addition 'dblClick' is called 
    """
        
    _height      = 72 
    _item_height = 23

    def __init__(self, *args, 
                 options = [],
                 doubleClick = None,  
                 autoHeight = False,                    # height dependend on n items, height must be tuple
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._options_getter = options
        self._options : list = None
        self._doubleClick_setter = doubleClick if callable (doubleClick) else None
        self._autoHeight = autoHeight

        self._get_properties ()

        # put into grid / layout 
        self._layout_add ()

        self._set_Qwidget_static ()
        self._set_Qwidget ()

        # connect signals 
        self.itemClicked.connect(self._on_selected)
        self.itemDoubleClicked.connect (self._on_doubleClick)


    @override
    def _get_properties (self): 

        super()._get_properties ()

        self._options = self._get_value (self._options_getter)
        self._val = str(self._val) if self._val is not None else None


    @override
    def _set_Qwidget (self, **kwargs):
        """ set value and properties of self Qwidget"""

        super()._set_Qwidget (**kwargs)

        # set items of listbox 

        self.clear()                                
        for item_text in self._options:
            item = QListWidgetItem (item_text, self)
            item.setSizeHint (QSize (0, self._item_height))

            # adapt foreground color of list items 
            if not self.light_mode:
                color = QColor("#BBBBBB")                       # hack - item.foreground().color() doesn't work
                item.setForeground (color)

        # set current item if not disabled 
        if not (self._disabled or self._disable_in_refresh):
            try: 
                irow = self._options.index (self._val)
            except: 
                irow = 0 
            self.setCurrentRow (irow)

        # set height of listbox dynamically (if height is tuple) 
        if self._autoHeight:
            self._set_height (self._height)



    def _set_height (self, height ):
        """ set a dynamic height of Listbox depending on number of items - height must be tuple"""
        if height is None or not isinstance (height, tuple): 
            return 

        min_height = height[0]
        max_height = height[1]

        nitems = len(self._options) if len(self._options) else 1
        calc_height = nitems * self._item_height + 4 

        min_height = min_height if min_height else calc_height
        max_height = max_height if max_height else calc_height
        if min_height > max_height:
            min_height, max_height = max_height, min_height

        self.setMinimumHeight(min_height)
        self.setMaximumHeight(max_height)        


    def _on_selected (self, aItem : QListWidgetItem):
        """ slot clicked"""
        self._set_value (aItem.text())


    def _on_doubleClick (self, aItem : QListWidgetItem):
        """ slot signal double clicked"""
        if self._doubleClick_setter:
            QTimer.singleShot (10, self._doubleClick_setter)  




class SpaceC:
    def __init__ (self, layout : QGridLayout, col, width : int =20, stretch=1): 
        """ sets properties of a space column in a grid layout  """
        if isinstance (layout, QGridLayout):
            layout.setColumnMinimumWidth (col,width)
            layout.setColumnStretch (col,stretch)

class SpaceR:
    def __init__ (self, layout : QGridLayout, row, height : int = 10, stretch=1): 
        """ sets properties of a space row in a grid layout  """
        if isinstance (layout, QGridLayout):
            layout.setRowMinimumHeight (row,height)
            layout.setRowStretch (row,stretch)


# ------------------------------------------------------------------------------
# ------------ test functions - to activate  -----------------------------------
# ------------------------------------------------------------------------------


class Test_Widgets (QMainWindow):


    def __init__(self):
        super().__init__()

        self.setWindowTitle('Test all widgets')
        self.setMinimumSize(QSize(700, 500))

        self.disabled = False
        l = QGridLayout()

        r = 0
        Label  (l,r,0,get="Header")
        Label  (l,r,1,fontSize=size.HEADER, colSpan=2,get="This is my header in 2 columns")
        CheckBox (l,r,3,fontSize=size.HEADER, text="Header", width=(90, 120), disable=lambda: self.disabled, )
        r += 1
        Label  (l,r,0,get="Label",width=(90,None))
        Label  (l,r,1,get=lambda: f"Disabled: {str(self.disabled)}", disable=lambda:self.disabled, lab_disable=True)
        Label  (l,r,2,get=lambda: f"Good {self.str_val()}", style=style.GOOD, styleRole = QPalette.ColorRole.Window)
        Label  (l,r,3,get=lambda: f"Disabled: {str(self.disabled)}", style=self.style )
        r += 1
        Label  (l,r,0,get="Field")
        Field  (l,r,1,get="initial", set=self.set_str, width=(80, 120), disable=True) # style=style.ERROR
        Field  (l,r,2,get=self.str_val, set=self.set_str, disable=lambda: self.disabled)
        Field  (l,r,3,get="Error", set=self.set_str, width=80, style=self.style, disable=lambda: self.disabled)
        r += 1
        Field  (l,r,0,lab="Field with label", get="initial", set=self.set_str, width=(80, 120))
        r += 1
        Label  (l,r,0,get="FieldI")
        FieldI (l,r,1,get=15, set=self.set_int, lim=(0,100), unit="kg", step=1, specialText="Automatic", width=80, style=style.ERROR)
        FieldI (l,r,2,get=self.int_val, set=self.set_int, lim=(1,100), step=1, disable=lambda: self.disabled, width=(80, 100), style=style.GOOD)
        FieldI (l,r,3,get=self.int_val, set=self.set_int, lim=(1,100), step=1, disable=lambda: self.disabled, style=self.style, width=(80, 100))
        r += 1
        Label  (l,r,0,get="FieldF")
        FieldF (l,r,1,get=-0.1234, set=self.set_float, width=80, lim=(-1,1), unit="m", step=0.1, dec=2, specialText="Automatic", disable=lambda: self.disabled)
        FieldF (l,r,2,get=self.float_val, set=self.set_float, lim=(1,100), dec=4, step=1.0, disable=lambda: self.disabled, style=style.GOOD)
        FieldF (l,r,3,get=self.float_val, set=self.set_float, lim=(1,100), width=80, dec=4, step=1.0, disable=True, style=self.style)
        r += 1
        FieldF (l,r,0, lab="FieldF combi",get=-0.1234, set=self.set_float, width=80, lim=(-1,1), unit="m", step=0.1, dec=2, specialText="Automatic", style=style.GOOD)
        r += 1
        Label  (l,r,0,get="Slider")
        Slider (l,r,1,get=-0.1234, set=self.set_float, width=80, lim=(-1,1))
        Slider (l,r,2,get=self.float_val, set=self.set_float, lim=(1,100), disable=lambda: self.disabled, style=style.GOOD)
        Slider (l,r,3,get=self.float_val()*100, set=self.set_float, lim=(100,1000), dec=0, width=80)
        r += 1
        Label  (l,r,0,get="ComboBox")
        ComboBox (l,r,1,options=["first","second"], set=self.set_str, width=80)
        ComboBox (l,r,2,options=["first","second"], set=self.set_str, disable=lambda: self.disabled)
        r += 1
        Label  (l,r,0,get="ComboSpinBox")
        ComboSpinBox (l,r,1,options=["first","second"], set=self.set_str, width=100)
        ComboSpinBox (l,r,2,options=["first","second"], set=self.set_str, disable=lambda: self.disabled)
        r += 1
        Label  (l,r,0,get="CheckBox")
        CheckBox (l,r,1,text="Check to disable", width=(90, 120), get=lambda: self.disabled, set=self.set_disabled)
        CheckBox (l,r,2,text="Slave check", get=self.bool_val, disable=lambda: self.disabled)
        r += 1
        Label  (l,r,0,get="SpaceR height=10")
        SpaceR (l,r, height=10)
        r += 1
        Label  (l,r,0,get="Button")
        Button (l,r,1,text="Toggle Disable", set=self.toggle_disabled, width=80)
        Button (l,r,2,text='Refresh', set=lambda: Widget.refresh_childs(self) )
        r += 1
        Label  (l,r,0,get="ToolButtons")
        l_tools = QHBoxLayout()
        l_tools.setSpacing (1)
        ToolButton (l_tools, icon=Icon.OPEN, set=self.toggle_disabled)
        ToolButton (l_tools, icon=Icon.SETTINGS, set=self.toggle_disabled)
        ToolButton (l_tools, icon=Icon.AE, set=self.toggle_disabled)
        l_tools.addStretch(2)
        l.addLayout (l_tools,r,1, 1, 2)

        l.setColumnStretch (0,1)
        l.setColumnStretch (1,1)
        l.setColumnStretch (2,2)
        l.setColumnStretch (4,1)

        l.setRowStretch (r+1,1)

        container = QWidget()
        container.setLayout (l) 
        # set_background (container, color="yellow",  alpha=0.5)
        self.setCentralWidget(container)

    def str_val (self):
        return ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))

    def int_val (self):
        return random.randrange(1, 100)
    
    def bool_val (self):
        return bool (random.randrange(0,1))
    
    def float_val (self):
        return random.uniform(0.0001, 99.9999)
    
    def style (self):
        if self.disabled:
            return style.NORMAL
        else: 
            return style.ERROR
    
    def set_str(self, aStr):
        if not isinstance (aStr, str): raise ValueError ("no string: ", aStr)
        print (f"Set string: {aStr}")
    def set_int(self, aInt):
        if not isinstance (aInt, int): raise ValueError ("no int: ", aInt)
        print (f"Set int: {aInt}")
    def set_float(self, aFloat):
        if not isinstance (aFloat, float): raise ValueError ("no int: ", aFloat)
        print (f"Set float: {aFloat}")
    def button_pressed(self):
        print (f"Button pressed")
    def toggle_disabled (self): 
        self.disabled = not self.disabled 
        Widget.refresh_childs (self) 
    def set_disabled (self, aBool): 
        self.disabled = aBool 
        Widget.refresh_childs (self) 




if __name__ == "__main__":

    from PyQt6.QtWidgets import QApplication, QMainWindow, QWidget, QGridLayout
    import random
    import string
    logging.basicConfig(level=logging.DEBUG)
 
    app = QApplication([])
    app.setStyle('fusion')
    # Strange: Without setStyleSheet, reset Widget.setPalette doesn't work .. !?
    # Segoe UI is the font of 'fusion' sttyle 
    # font = QFont ()
    # print (font.defaultFamily(), font.family(), font.families())
    app.setStyleSheet ("QWidget { font-family: 'Segoe UI' }")

    Test_Widgets().show()
    
    app.exec() 



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Proxies to Xoptfoil2 and Worker 

If the programs aren't in path-environment or in 'assets directory'
the path to program location location must be set 

   Worker.exePath = "xy/z/"
"""


import os
from tempfile       import NamedTemporaryFile
from glob           import glob
from pathlib        import Path

import time 
import shutil
import logging
import datetime 
import fnmatch

from subprocess     import Popen, run, PIPE
if os.name == 'nt':                                 # startupinfo only available in windows environment  
    from subprocess import STARTUPINFO, CREATE_NEW_CONSOLE, STARTF_USESHOWWINDOW, CREATE_NO_WINDOW


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


SW_NORMAL = 1 
SW_MINIMIZE = 6 

EXE_DIR_WIN    = 'assets/windows'                   # directory of exe files 
EXE_DIR_UNIX   = 'assets/linux'                    

TMP_INPUT_NAME = 'tmp~'                             # temporary input file (~1 will be appended)
TMP_INPUT_EXT  = '.inp'

#------- Helper function -----------------------------------------#

def is_younger (filePath, age_in_seconds): 
    """ returns true if filePathName is younger than age_in_seconds """

    younger = False 

    if os.path.isfile (filePath):
        now         = datetime.datetime.now()
        last_update = datetime.datetime.fromtimestamp(os.path.getmtime(filePath))
        tdelta = now - last_update
        seconds = tdelta.total_seconds()
        younger =  seconds < age_in_seconds

    return younger


def file_in_use (filePath):
    """ returns True if file is in use by another process"""
    
    in_use = False

    if os.path.exists(filePath):
        try:
            os.rename(filePath, filePath)
        except OSError as e:
            logger.warning (f"File {filePath} in use by another process")
            in_use = True 

    return in_use 



#------------------------------------------------------------------------------------

class X_Program:
    """ 
    Abstract superclass - Proxy to execute eg Xoptfoil2 and Worker 
    
        self will be executed in 'workingDir' which must be set 
            if it can't be extracted from airfoil path
    """

    NAME        = 'My_Program'
    NAME_EXE    = 'my_program'                             # stem of of exe file 

    version     = ''                                       # version of self - will be set in isReady
    exe_dir     = None                                     # where to find .exe 
    ready       = False                                    # is Worker ready to work 
    ready_msg   = ''                                       # ready or error message 


    def __init__ (self, workingDir : str = None):

        if workingDir and not os.path.isdir (workingDir):
            raise ValueError (f"Working directory '{workingDir}' does not exist" )

        self._workingDir        = workingDir                # directory in which self will be executed
        self._popen : Popen     = None                      # instance of subprocess when async
        self._returncode        = 0                         # returncode when async finished 
        self._pipe_error_lines  = None                      # errortext lines from stderr when finished
        self._pipe_out_lines    = []                        # output lines from stdout when finished

        self._tmp_inpFile       = None                      # tmpfile of this instance, to be deleted 


    def __repr__(self) -> str:
        """ nice representation of self """
        return f"<{type(self).__name__}>"


    @property
    def workingDir (self) -> str: 
        """ directory in which self will be executed"""
        return self._workingDir
    
   
    def isReady (self, project_dir : str, min_version : str = '') -> bool:
        """ 
        checks if self is available with min_version.

        Args: 
            project_dir: directory where there should be ./assets/... 
            min_version: check fpr min version number 
        """

        # ready already checked? 

        if self.ready: return True

        # find .exe

        version_ok = False
        ready_msg  = None 
        cls = self.__class__

        if self.exe_dir is None: 

            exe_dir, ready_msg = self._get_exe_dir (project_dir)

            if exe_dir is None:                                        # self not found anywhere
                cls.ready_msg = ready_msg
                logger.error (ready_msg)
                return 
            else:     
                cls.exe_dir = exe_dir
                logger.debug (ready_msg)

        # try to execute with -h help argument to get version 

        returncode = self._execute ('-h', workingDir=self._get_workingDir(), capture_output=True)


        # extract version and check version
        if returncode == 0 :
            for line in self._pipe_out_lines: 
                words = line.split()

                # first word is program name 
                if len (words) > 1 and words[0] == self.NAME:

                    # last word is version - compare single version numbers
                    cls.version = words[-1]
                    version_ok = True
                    min_nums = min_version.split(".") 

                    if len (min_nums[0]) > 0 :
                        cur_nums = self.version.split(".")
                        for i in range(len(min_nums)): 
                            try: 
                                cur_num = int(cur_nums[i])
                            except: 
                                cur_num = 0 
                            if cur_num < int(min_nums[i]):
                                version_ok = False
                    else: 
                        version_ok = False 
            
            if not version_ok:
                cls.ready_msg = f"wrong version {self.version} - need {min_version}"
            else: 
                cls.ready = True
                cls.ready_msg = f"Ready"

        else: 
            cls.ready_msg = f"{self.NAME_EXE} couldn't be executed"      

        if self.ready: 
            logger.info (f"{self.NAME} {self.version} {self.ready_msg}  (loading from: {self.exe_dir})" )
        else: 
            logger.error (f"{self.ready_msg}  (loading from: {self.exe_dir})" )

        return self.ready
    

    def isRunning (self) -> bool:
        """ still running when async - otherwise False"""

        isRunning = False
        if self._popen is not None: 

            self._popen.poll ()                 # check return code

            if self._popen.returncode is None: 

                isRunning = True

            else: 

                self._returncode       = self._popen.returncode

                new_lines = self._popen.stderr.readlines() if self._popen.stderr else []
                new_lines = [x.rstrip() for x in new_lines]                 # remove \r
                self._pipe_error_lines.extend(new_lines)
                
                new_lines = self._popen.stdout.readlines() if self._popen.stdout else []
                new_lines = [x.rstrip() for x in new_lines]
                self._pipe_out_lines.extend(new_lines)

                if self._returncode:
                    logger.error (f"... {self.NAME_EXE} returncode: {self._returncode} - {'\n'.join (self._pipe_error_lines)}")

                    # put minimum info in error_lines if it isn't standard error return code of Xoptfoil
                    if self._returncode > 1: 
                        self._pipe_error_lines = [f"Process error: {self._returncode}"]

                else: 
                    logger.debug (f"... {self.NAME_EXE} finished {'\n'.join (self._pipe_out_lines)}")
                 
                self._popen = None              # close down process instance 

        return isRunning


    def remove_tmp_file (self, iretry = 0):
        """ os remove of temporary input file"""

        if self._tmp_inpFile and os.path.isfile (self._tmp_inpFile):       # remove tmpfile of worker 
            try: 

                os.remove(self._tmp_inpFile) 
                self._tmp_inpFile = None 

            except OSError as exc: 
                if iretry < 5: 
                    iretry +=1
                    logger.warning (f"{self} Could not delete tmp input file '{self._tmp_inpFile}' - Retry {iretry}")
                    time.sleep (0.1)
                    self.remove_tmp_file (iretry=iretry)
                else: 
                    logger.error (f"Could not delete tmp input file '{self._tmp_inpFile}' - {exc}")
        else:
            self._tmp_inpFile = None


    @property
    def finished_returncode (self): 
        """ returncode of subprocess when finished"""
        return self._returncode


    @property
    def finished_errortext (self): 
        """ errortext from subprocess in case returncode !=0 """

        # scan stderr and stdout for a line with 'Error: ' 

        text = None
        line : str

        if self._pipe_error_lines:
            for line in self._pipe_error_lines:
                _,_,text = line.partition ("Error: ")
                if text: return text 

            # if process was aborted no piped error text from xoptfoil2 - but a minimum message 
            return self._pipe_error_lines[0]

        if self._pipe_out_lines:
            for line in self._pipe_out_lines:
                _,_,text = line.partition ("Error: ")
                if text: return text 

        return text


    def finalize (self):
        """ do cleanup actions """

        self.remove_tmp_file ()


    def terminate (self):
        """ terminate os process of self"""

        if self.isRunning():
            self._popen.terminate()
            self._popen = None 


    # ---------- Private --------------------------------------

    def _execute (self, args = [], workingDir = None, capture_output : bool =False):
        """sync execute self in workingDir 

        Args:
            args: arguments of subprocess as list of strings
            capture_output: capture output in pipe. Defaults to False.
        Returns:
            returncode: = 0 if no error
        """

        if workingDir and not os.path.isdir (workingDir):
            returncode = 1
            logger.error (f"Working directory '{workingDir}' does not exist" )
            return returncode

        returncode  = 0 
        self._pipe_out_lines    = []                        # output lines from stderr when finished
        self._pipe_error_lines  = []                        # errortext lines from stdout when finished

        # build list of args needed by subprocess.run 

        exe = os.path.join (self.exe_dir, self.NAME_EXE) 

        if isinstance (args, list):
            arg_list = [exe] + args
        elif isinstance (args, str):
            arg_list = [exe] + [args]
        else: 
            arg_list = [exe]

        try:

            # uses subproocess run which returns a completed process instance 

            curDir = os.getcwd()
            if workingDir:
                os.chdir (workingDir)

            if capture_output:
                # needed when running as pyinstaller .exe 
                # https://stackoverflow.com/questions/7006238/how-do-i-hide-the-console-when-i-use-os-system-or-subprocess-call/7006424#7006424
                if os.name == 'nt':
                    flags = CREATE_NO_WINDOW    
                else: 
                    flags  = 0                      # posix must be 0       

            logger.info (f"... {self.NAME_EXE} run sync: '{args}' in: {workingDir}")

            process = run (arg_list, text=True, capture_output=capture_output, creationflags=flags)

            returncode  = process.returncode

            if returncode:
                logger.error (f"... {self.NAME_EXE} ended: '{process}'")

            # finished - nice output strings 

            if process.stderr:  
                self._pipe_error_lines = process.stderr.split ("\n")
                logger.error (f"... {self.NAME_EXE} stderr: {"\n".join (self._pipe_error_lines)}")

                if self._pipe_error_lines[0] == "STOP 1":
                    # the error message will be in stdout
                    self._pipe_error_lines = []

            if capture_output and process.stdout: 
                self._pipe_out_lines = process.stdout.split ("\n")
                # logger.debug (f"... {self.NAME_EXE} stdout: {"\n".join (self._pipe_out_lines)}")

        except FileNotFoundError as exc:

            returncode = 1
            self._pipe_error_lines = str(exc)

            logger.error (f"... exception {self.NAME_EXE}: {exc}")

        finally: 

            os.chdir (curDir)

        return  returncode


    def _execute_async (self, args = [], workingDir = None, capture_output : bool =False):
        """async execute self in workingDir 

        Args:
            args: arguments of subprocess as list of strings
            capture_output: capture output in pipe. Defaults to False.
        Returns:
            returncode: = 0 if no error
        """

        if workingDir and not os.path.isdir (workingDir):
            raise ValueError (f"Working directory '{workingDir}' does not exist" )

        returncode  = 0 
        self._pipe_out_lines    = []                        # errortext lines from stderr when finished
        self._pipe_error_lines  = []                        # output lines from stdout when finished

        # build list of args needed by subprocess.run 

        exe = os.path.join (self.exe_dir, self.NAME_EXE) 

        if isinstance (args, list):
            arg_list = [exe] + args
        elif isinstance (args, str):
            arg_list = [exe] + [args]
        else: 
            arg_list = [exe]
        # run either sync or async 

        try:

            # uses subproccess Popen instance to start a subprocess

            curDir = os.getcwd()
            if workingDir:
                os.chdir (workingDir)

            if capture_output:
                stdout = PIPE                               # output is piped to suppress window 
                stderr = PIPE                               # Xoptfoil will write error to stderr
                # needed when running as pyinstaller .exe 
                # https://stackoverflow.com/questions/7006238/how-do-i-hide-the-console-when-i-use-os-system-or-subprocess-call/7006424#7006424
                if os.name == 'nt':
                    flags = CREATE_NO_WINDOW    
                else: 
                    flags  = 0                              # posix must be 0 
                startupinfo = self._get_popen_startupinfo (SW_NORMAL)  
            else: 
                stdout = None 
                stderr = PIPE                               # Xoptfoil will write error to stderr
                if os.name == 'nt':
                    flags  = CREATE_NEW_CONSOLE             # a new console is created (Xoptfoil) 
                else: 
                    flags  = 0                              # posix must be 0 
                startupinfo = self._get_popen_startupinfo (SW_MINIMIZE)  

            logger.info (f"... {self.NAME_EXE} run async: '{args}' in: {workingDir}")

            popen = Popen (arg_list, creationflags=flags, text=True, **startupinfo, 
                                stdout=stdout, stderr=stderr)  

            popen.poll()                            # update returncode

            returncode  = popen.returncode if popen.returncode is not None else 0   # async returns None 

            if returncode:
                logger.error (f"... {self.NAME_EXE} ended: '{popen}'")

                self._pipe_error_lines = popen.stderr.readlines()
                logger.error (f"... {self.NAME_EXE} stderr: {"\n".join (self._pipe_error_lines)}")

            # keep for later poll 
            self._popen = popen 

        except FileNotFoundError as exc:

            returncode = 1
            self._pipe_error_lines = str(exc)

            logger.error (f"... exception {self.NAME_EXE}: {exc}")

        finally: 

            os.chdir (curDir)

        return  returncode



    def _get_popen_startupinfo (self, show : int):
        """ returns popen startinfo parm to eg. minimize shell window - only windows"""

        if os.name == 'nt':
            if show != SW_NORMAL:
                startupinfo = STARTUPINFO()  
                startupinfo.dwFlags |= STARTF_USESHOWWINDOW       
                startupinfo.wShowWindow = show
                return dict(startupinfo=startupinfo)
        return dict(startupinfo=None) 


    def _get_exe_dir (self, project_dir : str): 
        """
        trys to find path to call programName
        
        If found, returns exePath and ready_msg
        If not, return None and ready_msg (error)"""

        exe_dir  = None
        ready_msg = None 

        if os.name == 'nt':
            assets_dir = EXE_DIR_WIN
        else: 
            assets_dir = EXE_DIR_UNIX  

        assets_dir = os.path.normpath (assets_dir)  
        check_dir1 = os.path.join (project_dir , assets_dir)                            # .\modules\assets\...
        check_dir2 = os.path.join (os.path.dirname (project_dir), assets_dir)           # .\assets\...

        if shutil.which (self.NAME_EXE, path=check_dir1) : 
            exe_dir  = os.path.abspath(check_dir1) 
            ready_msg = f"{self.NAME_EXE} found in: {exe_dir}"
        elif shutil.which (self.NAME_EXE, path=check_dir2) : 
            exe_dir  = os.path.abspath(check_dir2) 
            ready_msg = f"{self.NAME_EXE} found in: {exe_dir}"
        else: 
            exe_path = shutil.which (self.NAME_EXE)  
            if exe_path: 
                exe_dir = os.path.dirname (exe_path)
                ready_msg = f"{self.NAME_EXE} using OS search path to execute: {exe_dir}"
            else: 
                ready_msg = f"{self.NAME_EXE} not found either in '{assets_dir}' nor via OS search path" 
        return exe_dir, ready_msg


    def _get_workingDir (self, airfoil_path : str = None) -> str:
        """ 
        returns the actual working dir in which X_Program will be executed
            A workingDir set for self will have presetence, otherwise it is
            retrieved from airfoil_path or os current working dir is taken  
        """

        if self.workingDir:
            return self.workingDir
        elif airfoil_path and os.path.isfile (airfoil_path) : # and os.path.isabs(airfoil_path)
            return os.path.dirname (airfoil_path)
        else:
            return os.getcwd()



# ------------------------------------------------------------



class Xoptfoil2 (X_Program):
    """ 
    Proxy to execute Xoptfoil2
    
        self will be executed in 'workingDir' which must be set if is not current dir
        The 'inputfile' must be in 'workingDir' 
    """

    NAME        = 'Xoptfoil2'
    NAME_EXE    = 'xoptfoil2'                               # stem of of exe file 

    RUN_CONTROL = 'run_control'                             # file name of control file 
    STILL_ALIVE = 10                                        # max. age in seconds of run_control

    RESULT_DIR_POSTFIX = '_temp'                            # result directory of Xoptfoil2 postfix of 'outname'

    @staticmethod
    def remove_resultDir (airfoil_pathFileName : str, only_if_older = False):
        """ 
        deletes the Xoptfoil2 result directory 
        """ 

        resultDir = str(Path(airfoil_pathFileName).with_suffix('')) + Xoptfoil2.RESULT_DIR_POSTFIX
        shutil.rmtree(resultDir, ignore_errors=True)


    @property
    def run_control_filePath (self):
        """ returns filePath of run_control"""
        if self.workingDir: 
            return os.path.join(self.workingDir, Xoptfoil2.RUN_CONTROL)
        else:
            return Xoptfoil2.RUN_CONTROL


    def run (self, outname:str, input_file:str=None, seed_airfoil:str =None):
        """ run self async in self workingDir

        Args:
            outname: output name for generated airfoil
            inputfile: name of input file. Defaults to 'outname'.inp.
            seed_airfoil: optional seed airfoil filename.
        Returns: 
            returncode: = 0 - no errors (which could be retrieved via 'finished_errortext' )
        """

        args = []
        if seed_airfoil  : args.extend(['-a', seed_airfoil])
        if outname       : args.extend(['-o', outname])  
        
        if input_file is None: input_file = outname + '.inp'
        args.extend(['-i', input_file]) 

        # add 'mode' option - will write error to stderr
        args.extend(['-m', 'child']) 

        returncode = self._execute_async (args=args, workingDir=self.workingDir)

        return returncode


    def isRunning (self) -> bool:
        """ 
        - still running?    ... process when async
                            ... otherwise check run_control file if self was started from outside 
        """

        if self._popen:                         # started program myself as process
            running = super().isRunning ()
        else:                                   # Xoptfoil was started from outside 
            running = is_younger (self.run_control_filePath,  Xoptfoil2.STILL_ALIVE)

        if not running and os.path.isfile (self.run_control_filePath):
            # remove old run_control 
            os.remove (self.run_control_filePath)
 
        return running 



    def get_progress (self):
        """ returns no of steps, no of designs and objective function when running otherwise 0,0, 1.0 """

        # format of run_control 
            # !stop
            # !run-info; step: 3; design: 3; fmin:  0.9919478

        steps   = 0 
        designs = 0 
        objFun  = 1.0 
        lines = []

        if os.path.isfile (self.run_control_filePath):
            with open(self.run_control_filePath, 'r') as file:
                lines = file.readlines()
                file.close()

        for line in lines: 
            infos = line.split(";")
            if len(infos) == 4: 
                try: 
                    steps   = int(infos[1].split(":")[1])
                    designs = int(infos[2].split(":")[1])
                    objFun  = float(infos[3].split(":")[1])
                except: 
                    pass

        return steps, designs, objFun


    def stop (self):
        """
        tries to stop self with a 'stop' command in 'run_control' file 
            self must run in 'workingDir'
        """

        with open(self.run_control_filePath, 'w+') as file:
            file.write("stop")
            file.close()


# ------------------------------------------------------------



class Worker (X_Program):
    """ proxy to execute Worker commands"""

    NAME        = 'Worker'
    NAME_EXE    = 'worker'                             # stem of of exe file 


    # -- static methods --------------------------------------------

    @staticmethod
    def polarDir (airfoil_pathFileName : str) -> str:
        """ returns polar directory of airfoil having airfoil_pathFileName"""

        return str(Path(airfoil_pathFileName).with_suffix('')) + '_polars'

    @staticmethod
    def flapped_suffix (flap_angle:float, x_flap:float, y_flap:float, y_flap_spec:str) -> str:
        """ 
        name extension for flapped airfoil or polar file 
            '_f5.1' for defaults or 
            '_f-1.4_xf0.72_yf0.5_yspecYC' for non default values
        """
        if flap_angle == 0.0: return ''

        s_flap_angle = f"_f{flap_angle:.2f}".rstrip('0').rstrip('.') 
        s_x_flap     = f"_xf{x_flap:.2f}".rstrip('0').rstrip('.') if x_flap != 0.75 else ''
        s_y_flap     = f"_yf{y_flap:.2f}".rstrip('0').rstrip('.') if y_flap != 0.0  else ''
        s_y_spec     = f"_yspecYC" if y_flap_spec == 'y/c'  else ''

        return  f"{s_flap_angle}{s_x_flap}{s_y_flap}{s_y_spec}"


    @staticmethod
    def remove_polarDir (airfoil_pathFileName : str, only_if_older = False):
        """ 
        deletes polar directory 
        If only_if_older the directory is not removed if it is younger than of airfoilPathFileName
        """ 

        polarDir = Worker.polarDir (airfoil_pathFileName) if airfoil_pathFileName else None

        # sanity check 
        if not os.path.isdir(polarDir): return 

        remove = True 

        if only_if_older:
            if os.path.isfile(airfoil_pathFileName):

                # compare datetime of airfoil file and polar dir 
                ts = os.path.getmtime(polarDir)                 # file modification timestamp of a file
                polarDir_dt = datetime.datetime.fromtimestamp(ts)        # convert timestamp into DateTime object

                ts = os.path.getmtime(airfoil_pathFileName)              # file modification timestamp of a file
                airfoil_dt = datetime.datetime.fromtimestamp(ts)         # convert timestamp into DateTime object

                # add safety seconds (async stuff?) 
                if (airfoil_dt < (polarDir_dt + datetime.timedelta(seconds=2))):
                    remove = False 
            else: 
                remove = False 

        if remove: 
            shutil.rmtree(polarDir, ignore_errors=True)


    @staticmethod
    def get_existingPolarFile (airfoil_pathFileName, 
                               polarType : str, re : float, ma : float, ncrit : float,
                               flap_angle : float, x_flap : float, y_flap : float, y_flap_spec : str) -> str:
        """ 
        Get pathFileName of polar file if it exists 
        """      

        def parm_is_ok (id:str, val : float|None, decimals, args :list[str]) -> bool:
            """ inner func: check if paramter is in argunents of fileName"""
            if val is not None and decimals is not None: 
                val = round (val, decimals)

            for arg in args:
                if id == arg [:len(id)]:
                    if val is not None:
                        arg_val = arg[len(id):] if isinstance (val, str) else float(arg[len(id):])
                        if val == arg_val:                  # right arg 
                            return True                     #   right value
                        else:
                            return False                    #   wrong value
                    else: 
                        return False                        # this arg is too much
            
            if val is None: 
                return True                                 # this arg is correctly not there 
            else:
                return False                                # this arg is missing


        # remove a maybe older polarDir
        Worker.remove_polarDir (airfoil_pathFileName, only_if_older=True)    

        # build name of polar dir from airfoil file 
        polarDir = Worker.polarDir (airfoil_pathFileName)
        if os.path.isdir (polarDir):         

            fileNames = fnmatch.filter(os.listdir(polarDir), '*.txt')
            for fileName in fileNames:

                args = Path(fileName).stem.split('_')

                ok = True 

                # classic part 'T1_Re0.500_M0.00_N7.0'

                ok = ok and parm_is_ok ("Re", re/1000000, 3, args)
                ok = ok and parm_is_ok ("M",  ma, 2, args)
                ok = ok and parm_is_ok ("N",  ncrit, 1, args)
                ok = ok and parm_is_ok ("T",  int(polarType[1:]), 0, args)

                # flapped part '_f-1.4_xf0.72_yf0.5_yspecYC' for non default values

                ok = ok and parm_is_ok ("f", flap_angle, 1, args)

                x_flap_arg = None if x_flap ==0.75 else x_flap
                ok = ok and parm_is_ok ("xf", x_flap_arg, 2, args)

                y_flap_arg = None if y_flap ==0.0 else y_flap
                ok = ok and parm_is_ok ("yf", y_flap_arg, 2, args)

                y_flap_spec_arg = 'YC' if y_flap_spec =='y/c' else None
                ok = ok and parm_is_ok ("yspec", y_flap_spec_arg, None, args)

                if ok:
                    # logger.debug (f"<class Worker> found polar file {fileName} in {polarDir}")
                    return os.path.join (polarDir, fileName)        # return pathFileName

        logger.debug (f"<class Worker> No polar file in {polarDir}")
        return None


    #---------------------------------------------------------------


    def check_inputFile (self, inputFile=None):
        """ uses Worker to check an Xoptfoil2"""

        if not self.ready: return 1, self.NAME + " not ready"

        error_text = ""
        args = ['-w', 'check-input', '-i', inputFile]

        returncode = self._execute (args, capture_output=True)

        if returncode != 0:

            # worker output should something like ...
            #  Worker   -check-input jx-gt-10v3.inp
            #  - Processing input
            #    - Reading input jx-gt-10v3.inp
            #    - Output prefix jx-gt-10v3
            #  Error: max_speed should be between 0.001 and 0.5

            for line in self._pipe_out_lines:
                error_text = line.partition("Error:")[2].strip()
                if error_text != '': break 
            if error_text == '':
                raise ValueError ("Errortext not found in Workers")

        return returncode, error_text 


    def generate_polar (self, airfoil_pathFileName, 
                        polarType : str, 
                        re : float | list, 
                        ma : float | list, 
                        ncrit : float,
                        autoRange = True, spec = 'alpha', valRange= [-3, 12, 0.25], 
                        flap_angle : float | list = 0.0, x_flap=0.75, y_flap=0, y_flap_spec='y/t',
                        nPoints=None, run_async = True) -> int:
        """ 
        Generate polar for airfoilPathFileName in directory of airfoil.
        Returncode = 0 if successfully started (async) or finish (sync)
        """ 

        if not os.path.isfile(airfoil_pathFileName): 
            name = airfoil_pathFileName if len(airfoil_pathFileName) <= 40 else "..." + airfoil_pathFileName[-35:]
            raise ValueError (f"Airfoil '{name}' does not exist")

        if (polarType == 'T2'):
            polarTypeNo = 2
        else: 
            polarTypeNo = 1

        if spec == 'alpha':
            spec_al = True
        else: 
            spec_al = False

        if not isinstance (re, list): re = [re]
        if not isinstance (ma, list): ma = [ma]
        if flap_angle:
            if not isinstance (flap_angle, list): flap_angle = [flap_angle]
        else: 
            flap_angle=None

        # working directory in which self will execute 

        workingDir = self._get_workingDir (airfoil_pathFileName)

        # a temporary input file for polar generation is created

        self._tmp_inpFile = self._generate_polar_inputFile (workingDir, 
                                    re, ma, polarTypeNo, ncrit, autoRange, spec_al, valRange,
                                    flap_angles=flap_angle, x_flap=x_flap, y_flap=y_flap, y_flap_spec=y_flap_spec,
                                    nPoints=nPoints) 
        if not self._tmp_inpFile:
            raise RuntimeError (f"{self.NAME} polar generation failed: Couldn't create input file")

        # build args for worker - force outname as Worker wrongly uses airfoil name!

        # outname = Path(os.path.basename(airfoil_pathFileName)).stem
        outname = '' 

        args = self._build_worker_args ('polar-flapped',airfoil1=airfoil_pathFileName, inputfile=self._tmp_inpFile, outname=outname)

        # .execute either sync or async

        if run_async: 

            returncode = self._execute_async (args, capture_output=True, workingDir=workingDir)

        else:

            returncode = self._execute       (args, capture_output=True, workingDir=workingDir)

            self.remove_tmp_file ()         
        
        if returncode: 
            raise RuntimeError (f"Worker polar generation failed for {airfoil_pathFileName}")
            



    def set_flap (self, airfoil_fileName, 
                        x_flap : float = 0.75,
                        y_flap : float = 0.0,
                        y_flap_spec : str = 'y/t',
                        flap_angle : float | list = 0.0,
                        outname : str =None) -> int:
        """ 
        Set flap for airfoilPathFileName in directory of airfoil.

        Returns fileName of flapped airfoil in working Dir  
        """ 

        # sanity check airfoil_file 

        if self.workingDir:
            airfoil_pathFileName_abs = os.path.join(self.workingDir, airfoil_fileName)
        else: 
            airfoil_pathFileName_abs = airfoil_fileName

        if not os.path.isfile(airfoil_pathFileName_abs): 
            name = airfoil_pathFileName_abs if len(airfoil_pathFileName_abs) <= 40 else "..." + airfoil_pathFileName_abs[-35:]
            raise ValueError (f"Airfoil '{name}' does not exist")

        # name of the generated airfoil needed

        if not outname: 
            outname=f"{Path(airfoil_fileName).stem}{Worker.flapped_suffix (flap_angle, x_flap, y_flap, y_flap_spec)}"

        # working directory in which self will execute 

        workingDir = self._get_workingDir (airfoil_fileName)

        # a temporary input file for polar generation is created

        self._tmp_inpFile = self._generate_flap_inputFile (workingDir, x_flap, y_flap, y_flap_spec, flap_angle)         
        if not self._tmp_inpFile:
            raise RuntimeError (f"{self.NAME} setting flap failed: Couldn't create input file")

        # build args for worker 

        args = self._build_worker_args ('flap', airfoil1=airfoil_fileName, 
                                                inputfile=self._tmp_inpFile,
                                                outname=outname)
        # .execute  sync

        rc = self._execute (args, capture_output=True, workingDir=workingDir)

        if rc: 
            raise RuntimeError (f"Worker set flap failed: {self.finished_errortext}")
        else: 
            self.remove_tmp_file ()         

            flapped_fileName = outname + ".dat"
            # check if flapped airfoil was created  
            if os.path.isfile (os.path.join (workingDir, flapped_fileName)):
                return flapped_fileName
            else: 
                return None 


    def clean_workingDir (self, workingDir):
        """ 
        deletes temporary (older) files Worker creates in workingDir
        """ 
        if os.path.isdir(workingDir):

            # remove tmp input files of polar generation 
            match_path = os.path.join (workingDir, f"{TMP_INPUT_NAME}*{TMP_INPUT_EXT}")
            
            for f in glob(match_path):
                os.remove(f)


# ---------- Private --------------------------------------



    def _build_worker_args (self, action, actionArg='', airfoil1='', airfoil2='', outname='', inputfile=''):
        """ 
        return worker args as list of strings  
        """

        airfoil1_fileName = os.path.basename(airfoil1)
        airfoil2_fileName = os.path.basename(airfoil2)

        # info inputfile is in a dir - strip dir from path - local execution 

        local_inputfile = os.path.basename(inputfile) if inputfile else ''

        args = []

        if (action != ''): 
            if (action == 'help'): 
                args.extend(['-h'])
            else:                  
                args.extend(['-w', action])
                if (actionArg): args.extend([actionArg]) 
        else: 
            raise ValueError ('action for worker is mandatory')

        if (airfoil1 ): args.extend(['-a',  airfoil1_fileName])
        if (airfoil2 ): args.extend(['-a2', airfoil2_fileName])
        if (outname  ): args.extend(['-o',  outname]) 
        if (inputfile): args.extend(['-i',  local_inputfile])

        # add 'mode' option - will write error to stderr
        args.extend(['-m', 'child']) 

        return  args



    def _generate_polar_inputFile (self, 
                                  workingDir : str,         # if none self._workingDir is taken 
                                  reNumbers : list[float], maNumbers : list[float],
                                  polarType : int, ncrit : float,  
                                  autoRange : bool, spec_al: bool, valRange: list[float], 
                                  nPoints = None,
                                  flap_angles : list[float] = None, x_flap=None, y_flap=None, y_flap_spec=None,
                                   ) -> str:
        """ Generate a temporary polar input file for worker like this 

        &polar_generation
            generate_polars = .true.
            polar_reynolds  = 230000, 300000, 400000
            polar_mach      = 0.0, 0.2, 0.5
            type_of_polar = 1
            auto_Range = .true.
            op_mode = 'spec-al'
            op_point_range = -2.6, 11.0, 0.5
        /
        &xfoil_run_options
            ncrit = 7.0
        /
        &operating_conditions                          ! options to describe the optimization task
        x_flap                 = 0.75                  ! chord position of flap 
        y_flap                 = 0.0                   ! vertical hinge position 
        y_flap_spec            = 'y/c'                 ! ... in chord unit or 'y/t' relative to height
        flap_angle             = 0.0                   ! list of flap angles to be set


        Returns: 
            pathFilename of input file  """

        tmpFilePath = None

        # create tmp input file 

        with NamedTemporaryFile(mode="w", delete=False, dir=workingDir, prefix=TMP_INPUT_NAME, suffix=TMP_INPUT_EXT) as tmp:

            tmp.write ("&polar_generation\n")
            tmp.write ("  type_of_polar = %d\n" % polarType)  
            tmp.write ("  polar_reynolds  = %s\n" % (', '.join(str(e) for e in reNumbers))) 
            if max(maNumbers) > 0.0: 
                tmp.write ("  polar_mach  = %s\n" % (', '.join(str(e) for e in maNumbers))) 
            if autoRange:
                tmp.write ("  auto_range = .true.\n") 
                if valRange[2]:                                 # write only increment
                    tmp.write ("  op_point_range = , , %.2f \n" % (valRange[2])) 
            else:
                if spec_al:  tmp.write ("  op_mode = 'spec-al'\n") 
                else:        tmp.write ("  op_mode = 'spec-cl'\n") 
                tmp.write ("  op_point_range = %.2f , %.2f , %.2f \n" % (valRange[0], valRange[1], valRange[2])) 
            tmp.write ("/\n")

            tmp.write ("&xfoil_run_options\n")
            tmp.write ("  ncrit = %.1f\n" % ncrit)  
            tmp.write ("/\n")

            if flap_angles:
                tmp.write ("&operating_conditions\n")
                tmp.write (f"  x_flap = {x_flap}\n")  
                tmp.write (f"  y_flap = {y_flap}\n")  
                tmp.write (f"  y_flap_spec = '{y_flap_spec}'\n")  
                tmp.write (f"  flap_angle = {', '.join(str(f) for f in flap_angles)}\n") 
                tmp.write ("/\n")

            if nPoints is not None: 
                tmp.write ("&paneling_options\n")
                tmp.write (f"  npoint = {int(nPoints)}\n")  
                tmp.write ("/\n")

            tmpFilePath = tmp.name

        return tmpFilePath              



    def _generate_flap_inputFile (self, 
                                 workingDir : str = None,         # if none self._workingDir is taken 
                                  x_flap : float = 0.75,
                                  y_flap : float = 0.0,
                                  y_flap_spec : str = 'y/c',
                                  flap_angle : float | list = 0.0, 
                                  ) -> str:
        """ Generate a temporary polar input file for worker inworkingDir

        &operating_conditions                              ! options to describe the optimization task
            x_flap                 = 0.75                  ! chord position of flap 
            y_flap                 = 0.0                   ! vertical hinge position 
            y_flap_spec            = 'y/c'                 ! ... in chord unit or 'y/t' relative to height
            flap_angle             = 0.0                   ! list of flap angles to be applied        
        /

        Returns: 
            pathFilename of input file  """

        tmpFilePath = None

        flap_angles = flap_angle if isinstance (flap_angle, list) else [flap_angle]
        y_flap_spec = 'y/t' if y_flap_spec == 'y/t' else 'y/c'

        # create tmp input file 

        with NamedTemporaryFile(mode="w", delete=False, dir=workingDir, prefix=TMP_INPUT_NAME, suffix=TMP_INPUT_EXT) as tmp:

            tmp.write ("&operating_conditions\n")
            tmp.write (f"  x_flap = {x_flap:.2f}\n")  
            tmp.write (f"  y_flap = {y_flap:.2f}\n")  
            tmp.write (f"  y_flap_spec = '{y_flap_spec}'\n")  
            tmp.write (f"  flap_angle = {','.join(map(str, flap_angles))}\n")  
            tmp.write ("/\n")

            tmpFilePath = tmp.name

        return tmpFilePath              


# -------------- End --------------------------------------




# Main program for testing 
if __name__ == "__main__":

    # init logging 
    from base.common_utils      import init_logging
    init_logging (level= logging.DEBUG)


    Worker().isReady (project_dir="..\\..", min_version='1.0.3')

    if Worker.ready:

        worker = Worker()

        if os.path.isfile ('..\\..\\test_airfoils\\MH 30.dat'):
            airfoil = '..\\..\\test_airfoils\\MH 30.dat'
        elif os.path.isfile ('MH 30.dat'):
            airfoil = 'MH 30.dat'
        else: 
            logger.error (f"Airfoil file 'MH 30.dat' not found")
            exit()

        # build name of polar dir from airfoil file 
        polarDir = str(Path(airfoil).with_suffix('')) + '_polars'

        # ------- sync test ---------------------------------------------

        try: 
            worker.generate_polar (airfoil, 'T1', 700000, 0.0, 8.0, flap_angle=5.12, run_async=False)

            worker.generate_polar (airfoil, 'T1', 700000, 0.0, 8.0, run_async=False)

            logger.info ("\n".join (worker._pipe_out_lines))
            polar_file = worker.get_existingPolarFile (airfoil, 'T1', 700000, 0.0, 8.0, flap_angle=5.12)

            if polar_file:
                logger.info  (f"polar file found: {polar_file}")
            else: 
                logger.error (f"polar file not found")

            worker.finalize ()
            worker.remove_polarDir (airfoil)

        except ValueError as exc:
            logger.error (f"{exc}")
        except RuntimeError as exc:
            # logger.error (f"Polar failed: {exc}")
            logger.error (f"{worker}: {worker.finished_errortext}")

        

        # ------- async test ---------------------------------------------

        worker = Worker()

        try: 
            worker.generate_polar (airfoil, 'T1', 700000, 0.0, 8.0, run_async=True)

            secs = 0 
            while worker.isRunning ():
                time.sleep (0.5)
                secs += 0.5
                logger.debug (f"{worker} waiting: {secs}s")

            if worker.finished_returncode == 0:

                polar_file = worker.get_existingPolarFile (airfoil, 'T1', 700000, 0.0, 8.0)

                if polar_file:
                    logger.info  (f"polar file found: {polar_file}")
                else: 
                    logger.error (f"polar file not found")
            else: 
                logger.error (f"{worker}: {worker.finished_errortext}")

            worker.finalize ()
            worker.remove_polarDir (airfoil)

        except ValueError as exc:
            logger.error (f"{exc}")
        except RuntimeError as exc:
            # logger.error (f"Polar failed: {exc}")
            logger.error (f"{worker}: {worker.finished_errortext}")
